/*
 * Copyright (c) 1993,1994,1995,1997,1998
 *      Texas A&M University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Texas A&M University
 *      and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Developers:
 *     Russell Neeper, David K. Hess, Douglas Lee Schales, David R. Safford
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef __FreeBSD__
#include <stdlib.h>
#else
#include <malloc.h>
#endif
#include <memory.h>
#include "chario.h"
#include "macros.h"
#include "hosts.h"
#include "services.h"
#include "classes.h"
#include "dbfc.h"
#include "flags.h"

struct hosttable *hostlist = NULL;
struct hosttable *hostlist_tail = NULL;
static u_int32_t host_count = 0;
static u_int32_t net_count = 0;
static u_int32_t addr_count = 0;

struct hosttable *acclist = NULL;
struct hosttable *acclist_tail = NULL;
static struct accept *acceptlistbuf = NULL;
static u_int32_t accept_count = 0;

struct hosttable *rejlist = NULL;
struct hosttable *rejlist_tail = NULL;
static struct reject *rejectlistbuf = NULL;
static u_int32_t reject_count = 0;

static struct override *overridelistbuf;
static int override_count;
static int overridelistsize;

static PortListEntry tcprst_in[LIST_SIZE];
static PortListEntry tcprst_out[LIST_SIZE];


static unsigned int netmask(unsigned int);
static int endval(unsigned int, int);
static int numzeros(unsigned int);
static int maskbits(unsigned int);
static int compare_addr(struct accept *, struct accept *);
static int compare_override(struct override *, struct override *);
static struct hosttable *add_cidr(struct hosttable **, struct hosttable **,
				  unsigned int, u_int, u_int);
static u_int add_range(struct hosttable **, struct hosttable **,
		       unsigned int, unsigned int, int);

extern int parse_error;
extern int flag[MAX_FLAG];
extern int flag_count;

static struct hosttable *add_cidr(struct hosttable **, struct hosttable **,
				 u_int32_t, u_int, u_int);

/*-----------------------------------------------------------------------*/

static u_int32_t
netmask(unsigned int bits)
{
	return (0xFFFFFFFFUL << ((bits < 32) ? 32 - bits : 0));
}

static int
endval(u_int32_t addr, int bits)
{
        return addr + ~netmask(bits);
}

static int
numzeros(u_int32_t a)
{
	register int i;

	for (i=0; (i < 32) && !(a & 1); i++, a >>= 1);
	return i;
}       

static int
maskbits(u_int32_t a)
{
        register int i;

        i = 32 - numzeros(a);
        if ((a ^ netmask(i)) != 0) {
		fprintf(stderr, "Error: \"%s\", line %d: invalid netmask\n",
				getfilename(), getlinenum());
		exit(1);
	}
        return i;
}

/*-----------------------------------------------------------------------*/

void
init_hosts(void)
{
     override_count = 0;
     overridelistsize = sizeof(struct override);
     overridelistbuf = (struct override *)
	  		malloc(overridelistsize * sizeof(struct override));
     if (!overridelistbuf) {
	  fprintf(stderr, "%s, line %d: malloc failed\n",
		  __FILE__, __LINE__);
	  exit(1);
     }
}

/*-----------------------------------------------------------------------*/

static int
compare_addr(struct accept *x, struct accept *y)
{
	if (x->flag != y->flag)
		return x->flag - y->flag;
	if (x->bits != y->bits)
		return y->bits - x->bits;
	return x->ipaddr - y->ipaddr;
}

static int
compare_override(struct override *x, struct override *y)
{
	if (x->netmask != y->netmask)
		return x->netmask - y->netmask;
	return x->ipaddr - y->ipaddr;
}

void
writetables(char *file)
{
     int fd;
     struct hosttable *rove;
     unsigned int i, j;
     u_int32_t t;
     u_int8_t c1, c2;
     u_int16_t hte;
     unsigned short p;
     struct {
          unsigned char magic[4];
          unsigned char version[32];
 	  unsigned int networks_offset;
	  unsigned int classes_offset;
	  unsigned int accept_offset;
 	  unsigned int reject_offset;
	  unsigned int override_offset;
	  unsigned int tcpreset_offset;
	  unsigned int flags_offset;
     } header;


     if (accept_count > 0) {
     	  acceptlistbuf = (struct accept *)calloc(accept_count,
						  sizeof(struct accept));
	  if (acceptlistbuf == NULL) {
	       fprintf(stderr, "%s, line %d: malloc failed\n",
                       __FILE__, __LINE__);
	       exit(1);
	  }
	  for (i=0,rove=acclist; rove; i++,rove=rove->next) {
	       acceptlistbuf[i].ipaddr = rove->addr;
	       acceptlistbuf[i].bits = rove->bits;
	       acceptlistbuf[i].flag = rove->class;
	  }
	  qsort(acceptlistbuf, accept_count, sizeof(struct accept),
		(void *) compare_addr);
     }
     if (reject_count > 0) {
	  rejectlistbuf = (struct reject *)calloc(reject_count,
						  sizeof(struct reject));
	  if (rejectlistbuf == NULL) {
	       fprintf(stderr, "%s, line %d: malloc failed\n",
                       __FILE__, __LINE__);
	       exit(1);
	  }
	  for (i=0,rove=rejlist; rove; i++,rove=rove->next) {
	       rejectlistbuf[i].ipaddr = rove->addr;
	       rejectlistbuf[i].bits = rove->bits;
	       rejectlistbuf[i].flag = rove->class;
	  }
	  qsort(rejectlistbuf, reject_count, sizeof(struct reject),
		(void *) compare_addr);
     }


     if ((fd = open(file, O_WRONLY|O_CREAT|O_TRUNC, 0600)) == -1) {
	  perror(file);
          return;
     }
     /* seek to make room for the header which will be written last */
     lseek(fd, sizeof(header), SEEK_SET);


     printf("There %s %d network%s, %d host%s, and ",
          (net_count == 1) ? "is" : "are", net_count,
          (net_count == 1) ? "" : "s",
          host_count, (host_count == 1) ? "" : "s");


     /*
      * Write the list of hosts and cidr blocks
      */

     header.networks_offset = htonl(lseek(fd, 0, SEEK_CUR));

     t = htonl(addr_count);
     write(fd, &t, sizeof(t));
     for (rove=hostlist; rove; rove=rove->next) {
	  t = htonl(rove->addr);
	  write(fd, &t, sizeof(t));
	  write(fd, &rove->bits, sizeof(u_int8_t));
	  hte = htons(rove->class);
	  write(fd, &hte, sizeof(u_int16_t));
     }

     /*
      * Write the class tables
      */

     header.classes_offset = htonl(lseek(fd, 0, SEEK_CUR));

     writeclassestable(fd);

     /*
      * Write the accept table
      */

     header.accept_offset = htonl(lseek(fd, 0, SEEK_CUR));

     i = htonl(accept_count);
     write(fd, &i, sizeof(i));
     for(i=0; i<accept_count; i++) {
          t = htonl(acceptlistbuf[i].ipaddr);
          c1 = acceptlistbuf[i].bits;
          c2 = acceptlistbuf[i].flag;
          write(fd, &t, sizeof(t));
          write(fd, &c1, sizeof(c1));
          write(fd, &c2, sizeof(c2));
     }

     /*
      * Write the reject table
      */

     header.reject_offset = htonl(lseek(fd, 0, SEEK_CUR));

     i = htonl(reject_count);
     write(fd, &i, sizeof(i));
     for(i=0; i<reject_count; i++) {
          t = htonl(rejectlistbuf[i].ipaddr);
	  c1 = rejectlistbuf[i].bits;
	  c2 = rejectlistbuf[i].flag;
          write(fd, &t, sizeof(t));
          write(fd, &c1, sizeof(c1));
          write(fd, &c2, sizeof(c2));
     }

     /*
      * Write the override table
      */

     header.override_offset = htonl(lseek(fd, 0, SEEK_CUR));

     qsort(overridelistbuf, override_count, sizeof(struct override),
     		(void *) compare_override);
     i = htonl(override_count);
     write(fd, &i, sizeof(i));
     for(i=0; i<override_count; i++) {
          t = htonl(overridelistbuf[i].ipaddr);
          c1 = maskbits(overridelistbuf[i].netmask);
          write(fd, &t, sizeof(t));
          write(fd, &c1, sizeof(c1));
          for(j=0; j<LIST_SIZE; j++) {
               p = htons(overridelistbuf[i].ports[j].begin);
               write(fd, &p, sizeof(p));
               p = htons(overridelistbuf[i].ports[j].end);
               write(fd, &p, sizeof(p));
	       if (j == overridelistbuf[i].ports[0].begin)
	            break;
          }
     }

     /*
      * Write the tcpreset table
      */

     header.tcpreset_offset = htonl(lseek(fd, 0, SEEK_CUR));

     for (i=0; i<LIST_SIZE; i++) {
	  u_int16_t t[2];
	  t[0] = htons(tcprst_in[i].begin);
          t[1] = htons(tcprst_in[i].end);
	  write(fd, t, sizeof(t));
	  if (i == tcprst_in[0].begin)
	       break;
     }
     for (i=0; i<LIST_SIZE; i++) {
	  u_int16_t t[2];
	  t[0] = htons(tcprst_out[i].begin);
          t[1] = htons(tcprst_out[i].end);
	  write(fd, t, sizeof(t));
	  if (i == tcprst_in[0].begin)
	       break;
     }

     /*
      * Write the flags
      */

     header.flags_offset = htonl(lseek(fd, 0, SEEK_CUR));

     i = htonl(flag_count);
     write(fd, &i, sizeof(i));
     for (i=0; i<MAX_FLAG; i++) {
	  if (flag[i] != -1) {
	       c1=i;
	       c2=flag[i];
               write(fd, &c1, sizeof(c1));
               write(fd, &c2, sizeof(c2));
	  }
     }

     /*
      * Write the file header
      */
     lseek(fd, 0, SEEK_SET);
     strncpy(header.magic, "DBFC", 4);
     memset(header.version, 0, sizeof(header.version));
     strncpy(header.version, VERSION, sizeof(header.version)-1);
     write(fd, &header, sizeof(header));

     close(fd);
}

/*-----------------------------------------------------------------------*/

static u_int32_t *ipstack = NULL;
static int ipstack_index = 0;
static int ipstack_size = 0;

void
push_ip(u_int32_t ip)
{
	if (ipstack_index >= ipstack_size) {
		ipstack_size += 10;
		ipstack = (u_int32_t *)realloc(ipstack,
				ipstack_size * sizeof(u_int32_t));
		if (ipstack == NULL) {
			fprintf(stderr, "%s, line %d: realloc failed\n",
				__FILE__, __LINE__);
			exit(1);
		}
	}
	ipstack[ipstack_index++] = ip;
}

u_int32_t *
pop_ip()
{
	if (ipstack_index)
		return &ipstack[--ipstack_index];
	return NULL;
}

/*-----------------------------------------------------------------------*/
 
struct host_def *
check_host(u_int32_t host)
{
	register struct hosttable *hrove;

	for (hrove=hostlist; hrove; hrove=hrove->next)
		if ((hrove->addr == host) && (hrove->bits == 32))
			return hrove->def;
	return NULL;
}

/*-----------------------------------------------------------------------*/

static struct hosttable *
add_cidr(struct hosttable **head, struct hosttable **tail,
	 u_int32_t addr, u_int bits, u_int class)
{
	struct hosttable *hrove;

	hrove = (struct hosttable *)malloc(sizeof(struct hosttable));
	if (!hrove) {
		fprintf(stderr, "%s, line %d: malloc failed\n",
				__FILE__, __LINE__);
		exit(1);
	}

	if ((addr & netmask(bits)) != addr) {
		fprintf(stderr, "Warning: \"%s\", line %d: disregarding "
			"non-zero host portion of address\n",
			getfilename(), getlinenum());
	}
	hrove->addr = addr & netmask(bits);
	hrove->bits = bits;
	hrove->class = class;
	hrove->def = NULL;
	hrove->next = NULL;

	if (*tail == NULL)
		*head = hrove;
	else
		(*tail)->next = hrove;
	*tail = hrove;

	return hrove;
}

/*-----------------------------------------------------------------------*/

static u_int
add_range(struct hosttable **head, struct hosttable **tail,
	  u_int32_t addr1, u_int32_t addr2, int class)
{
        register int bits, roll=0;
	u_int count=0;

        while (!roll && (addr1 <= addr2)) {
                bits = 32 - numzeros(addr1);
                while (endval(addr1, bits) > addr2)
                        bits++;
		add_cidr(head, tail, addr1, bits, class);
		count++;

                addr1 = endval(addr1, bits) + 1;
                roll = (addr1) ? 0 : 1;
        }
	return count;
}

/*-----------------------------------------------------------------------*/

void
do_host(void)
{
	int class_entry;
	struct hosttable *hrove;
	u_int32_t *ip;

	ip = pop_ip();
	if (ip == NULL)
		return;

	class_entry = make_class();
	destroy_all_services();

	while (ip) {
		hrove = add_cidr(&hostlist, &hostlist_tail, *ip,
				 32, class_entry);
		addr_count++;

		if (hrove && (hrove->def == NULL)) {
			if (!(hrove->def = (struct host_def *)
			      malloc(sizeof(struct host_def)))) {
				fprintf(stderr, "%s, line %d: malloc failed\n",
						__FILE__, __LINE__);
				exit(1);
			}
			hrove->def->hfile = hrove->def->nfile = NULL;
			hrove->def->hline = hrove->def->nline = 0;
		}

		if (hrove->def->hfile)
			free(hrove->def->hfile);
		hrove->def->hfile = strdup(getfilename());
		hrove->def->hline = getlinenum();

		ip=pop_ip();
	}

	host_count++;
}

/*-----------------------------------------------------------------------*/

void
do_network(struct hostrange hr, u_int32_t mask, int bits)
{
	int class_entry;

	class_entry = make_class();
	destroy_all_services();

	if (hr.end != hr.begin) {
		addr_count += add_range(&hostlist, &hostlist_tail,
			  		hr.begin, hr.end, class_entry);
	} else {
		if (bits == -1)
			add_cidr(&hostlist, &hostlist_tail,
					 hr.begin, maskbits(mask), class_entry);
		else
			add_cidr(&hostlist, &hostlist_tail,
					 hr.begin, bits, class_entry);
		addr_count++;
	}
	net_count++;
}

/*-----------------------------------------------------------------------*/

void
do_tcpreset()
{
     PortListEntry t_src_in[LIST_SIZE], t_src_out[LIST_SIZE],
		   u_dst_in[LIST_SIZE], u_dst_out[LIST_SIZE],
		   u_src_in[LIST_SIZE], u_src_out[LIST_SIZE],
		   i_typ_in[LIST_SIZE], i_typ_out[LIST_SIZE];

     gen_class(tcprst_in, tcprst_out, t_src_in, t_src_out,
	       u_dst_in, u_dst_out, u_src_in, u_src_out,
	       i_typ_in, i_typ_out);
     destroy_all_services();

     if (t_src_in[0].begin || t_src_out[0].begin ||
	 u_dst_in[0].begin || u_dst_out[0].begin ||
	 u_src_in[0].begin || u_src_out[0].begin ||
	 i_typ_in[0].begin || i_typ_out[0].begin) {
	  fprintf(stderr, "INFO: \"%s\", line %d: 'tcpreset' only "
		 "supports 'in' and 'out' TCP destination ports\n",
		 getfilename(), getlinenum());
     }
}

/*-----------------------------------------------------------------------*/

void
add_accept(struct hostrange hr, u_int32_t mask, int bits, norminv flag)
{
	if (hr.end != hr.begin) {
		accept_count += add_range(&acclist, &acclist_tail,
			  		  hr.begin, hr.end, flag);
	} else {
		if (bits == -1)
			add_cidr(&acclist, &acclist_tail,
					hr.begin, maskbits(mask), flag);
		else
			add_cidr(&acclist, &acclist_tail,
					hr.begin, bits, flag);
		accept_count++;
	}
}

/*-----------------------------------------------------------------------*/

void
add_reject(struct hostrange hr, u_int32_t mask, int bits, norminv flag)
{
	if (hr.end != hr.begin) {
		reject_count += add_range(&rejlist, &rejlist_tail,
			  		  hr.begin, hr.end, flag);
	} else {
		if (bits == -1)
			add_cidr(&rejlist, &rejlist_tail,
					hr.begin, maskbits(mask), flag);
		else
			add_cidr(&rejlist, &rejlist_tail,
					hr.begin, bits, flag);
		reject_count++;
	}
}

/*-----------------------------------------------------------------------*/

void
add_override(u_int32_t ipaddr, u_int32_t mask, int bits)
{
     PortListEntry t_dst_in[LIST_SIZE], t_dst_out[LIST_SIZE],
     		   t_src_in[LIST_SIZE], t_src_out[LIST_SIZE],
		   u_dst_in[LIST_SIZE], u_dst_out[LIST_SIZE],
		   u_src_in[LIST_SIZE], u_src_out[LIST_SIZE],
		   i_typ_in[LIST_SIZE], i_typ_out[LIST_SIZE];

     if (bits == -1)
     	  maskbits(mask);
     else
	  mask = netmask(bits);

     if(override_count == overridelistsize){
	  overridelistsize *= 2;
	  overridelistbuf = (struct override *) realloc(overridelistbuf,
		       overridelistsize * sizeof(struct override));
	  if(!overridelistbuf){
	       fprintf(stderr, "%s, line %d: realloc failed\n",
		       __FILE__, __LINE__);
	       exit(1);
	  }
     }
     overridelistbuf[override_count].ipaddr = ipaddr & mask;
     overridelistbuf[override_count].netmask = mask;

     gen_class(t_dst_in, t_dst_out, t_src_in, t_src_out,
	       u_dst_in, u_dst_out, u_src_in, u_src_out,
	       i_typ_in, i_typ_out);
     destroy_all_services();

     if (t_dst_in[0].begin || t_src_in[0].begin ||
	 t_src_out[0].begin || u_dst_in[0].begin ||
	 u_dst_out[0].begin || u_src_in[0].begin ||
	 u_src_out[0].begin || i_typ_in[0].begin || i_typ_out[0].begin)
	  fprintf(stderr, "INFO: \"%s\", line %d: 'override' only "
			  "supports 'out' TCP ports\n",
		  	  getfilename(), getlinenum());

     memcpy(overridelistbuf[override_count].ports, t_dst_out,
			LIST_SIZE * sizeof(PortListEntry));

     override_count++;
}

/*-----------------------------------------------------------------------*/
