/*
 * Copyright (c) 1993,1994,1995,1997,1998
 *      Texas A&M University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Texas A&M University
 *      and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Developers:
 *     Russell Neeper, David K. Hess, Douglas Lee Schales, David R. Safford
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/param.h>

#include "drawbridge.h"
#include "dbmgr.h"
#include "tokens.h"
#include "misc.h"

#define __LOAD__
#include "text.h"

typedef struct {
	unsigned char magic[4];
	unsigned char version[32];
	unsigned int networks_offset;
	unsigned int classes_offset;
	unsigned int accept_offset;
	unsigned int reject_offset;
	unsigned int override_offset;
	unsigned int tcpreset_offset;
	unsigned int flags_offset;
} fileHeader;

static int loadAddrTable(char *, FILE *, u_int);
static int readPortList(FILE *, PortListEntry *);
static int loadClassTable(FILE *, int, u_int16_t, char *);
static int loadClassTables(char *, FILE *, u_int);
static int loadAccRejTable(char *, FILE *, u_int, int);
static int loadOverrideTable(char *, FILE *, u_int);
static int loadTCPresetTable(char *, FILE *, u_int);
static int loadFlags(char *, FILE *, u_int);
static int loadTables(FILE *, fileHeader, char *);


/*------------------------------------------------------------*/

void
usageLoad(int verbose)
{
	fprintf(theOutput, "%s%s",
		(verbose) ? loadText[0] : loadText[0]+1,
		(verbose) ? loadText[1] : "");
}

/*------------------------------------------------------------*/

static int
loadAddrTable(char *file, FILE *fp, u_int offset)
{
	u_int32_t count, loaded;
	addrTreeReq u;
	register u_long i;

	/*
	 * Seek to the networks table
	 */
	fseek(fp, offset, SEEK_SET);

	if (fread(&count, sizeof(u_long), 1, fp) < 1) {
		fprintf(stderr,"   Error reading from %s\n", file);
		return ERROR;
	}
	count = ntohl(count);
	loaded = count;

	u.newtable = 1;
	while (count) {
		in_addr_t addr;
		u_int8_t bits;
		u_int16_t class;

		for (i=0; count && (i < 128); i++, count--) {
			if ((fread(&addr, sizeof(addr), 1, fp) < 1) ||
			    (fread(&bits, sizeof(bits), 1, fp) < 1) ||
			    (fread(&class, sizeof(class), 1, fp) < 1)) {
				fprintf(stderr,"   Error reading from %s\n",
					file);
				return ERROR;
			}
			u.addr[i].s_addr = ntohl(addr.s_addr);
			u.bits[i] = bits;
			u.class[i] = ntohs(class);
		}
		u.num = i;
		if (ioctl(theSocket, DIOCSNETWORK, &u)) {
			dfm_perror("DIOCSNETWORK");
			return ERROR;
		}
		u.newtable = 0;
	}

	if (loaded > 0)
		fprintf(theOutput,"   Address table loaded\n");
	return NOERROR;
}

/*------------------------------------------------------------*/

static int
readPortList(FILE *fd, PortListEntry *p)
{
	register u_short i,n=0;

	for (i=0; i < LIST_SIZE; i++) {
		if (fread(p, sizeof(PortListEntry), 1, fd) != 1)
			return ERROR;
		p->begin = ntohs(p->begin);
		p->end = ntohs(p->end);
		if (i == 0)
			n = p->begin;
		if (i == n)
			break;
		p++;
	}
	return NOERROR;
}

/*------------------------------------------------------------*/

static int
loadClassTable(FILE *fp, int id, u_int16_t class, char *name)
{
	PortListEntry list[LIST_SIZE];

	memset(list, 0, sizeof(list));
	if (fp) {
		if (readPortList(fp, list)) {  
			fprintf(theErrors, "   Error reading '%s' "
				"access table\n", name);
			return ERROR;
		}
	}
	if (portListRequest(WRITE, id, class, list, LIST_TABLE_SIZE))
		return ERROR;
	return NOERROR;
}

/*------------------------------------------------------------*/

static int
loadClassTables(char *file, FILE *fp, u_int offset)
{
	u_int16_t r, num;

	/* Seek to the classes table */
	fseek(fp, offset, SEEK_SET);

	/* Read the number of classes */
	if (fread(&num, sizeof(num), 1, fp) != 1)
		return ERROR;
	num = ntohs(num);

	/* Load each class table */
	for (r=0; r<num; r++) {
		if (loadClassTable(fp, T_DST_IN, r, "tcp_dst_in") == ERROR)
			return ERROR;
		if (loadClassTable(fp, T_DST_OUT, r, "tcp_dst_out") == ERROR)
			return ERROR;
		if (loadClassTable(fp, T_SRC_IN, r, "tcp_src_in") == ERROR)
			return ERROR;
		if (loadClassTable(fp, T_SRC_OUT, r, "tcp_src_out") == ERROR)
			return ERROR;
		if (loadClassTable(fp, U_DST_IN, r, "udp_dst_in") == ERROR)
			return ERROR;
		if (loadClassTable(fp, U_DST_OUT, r, "udp_dst_out") == ERROR)
			return ERROR;
		if (loadClassTable(fp, U_SRC_IN, r, "udp_src_in") == ERROR)
			return ERROR;
		if (loadClassTable(fp, U_SRC_OUT, r, "udp_src_out") == ERROR)
			return ERROR;
		if (loadClassTable(fp, I_TYP_IN, r, "icmp_type_in") == ERROR)
			return ERROR;
		if (loadClassTable(fp, I_TYP_OUT, r, "icmp_type_out") == ERROR)
			return ERROR;
	}

	fprintf(theOutput,"   Class table loaded\n");
	return NOERROR;
}

/*------------------------------------------------------------*/

static int
loadAccRejTable(char *file, FILE *fp, u_int offset, int id)
{
	u_int i, count;
	rejaccTableReq req;
	RejAccTableEntry *table;
	in_addr_t n;
	u_int8_t b, f;

	/*
	 * Seek to the table and read the table size
	 */
	fseek(fp, offset, SEEK_SET);
	if (fread(&count, sizeof(count), 1, fp) < 1) {
		fprintf(stderr, "   Error reading from %s\n", file);
		return ERROR;
	}
	count = ntohl(count);

	if (count != 0) {
		table = (RejAccTableEntry *)
			calloc(count, sizeof(RejAccTableEntry));
		if (table == NULL) {
			fprintf(stderr, "   Unable to allocate memory "
					"for %s table\n", (id == T_REJTBL)
					? "reject" : "accept");
			return ERROR;
		}

		for (i=0; i < count; i++) {
			if ((fread(&n, sizeof(n), 1, fp) < 1) ||
			    (fread(&b, sizeof(b), 1, fp) < 1) ||
			    (fread(&f, sizeof(f), 1, fp) < 1)) {
				fprintf(stderr, "   Error reading from %s\n",
					file);
				free(table);
				return ERROR;
			}
			table[i].network.s_addr = ntohl(n.s_addr);
			table[i].bits = b;
			table[i].flag = f;
		}

		req.ptr = table;
		req.bytes = count * sizeof(RejAccTableEntry);
		if (id == T_REJTBL) {
			if (ioctl(theSocket, DIOCSREJECT, &req)) {
				dfm_perror("DIOCSREJECT");
				free(table);
				return ERROR;
			}
		} else {
			if (ioctl(theSocket, DIOCSACCEPT, &req)) {
				dfm_perror("DIOCSACCEPT");
				free(table);
				return ERROR;
			}
		}
		free(table);
	}

	if (count > 0)
		fprintf(theOutput, "   %s table loaded: %d entries\n",
			(id == T_REJTBL) ? "Reject" : "Accept", count);
	return NOERROR;
}

/*------------------------------------------------------------*/

static int
loadOverrideTable(char *file, FILE *fp, u_int offset)
{
	u_int i, j, k, count;
	overrideTableReq req;
	OverrideTableEntry *table;
	in_addr_t n;
	u_int8_t b;
	PortListEntry *p;

	/*
	 * Seek to the table and read the table size
	 */
	fseek(fp, offset, SEEK_SET);
	if (fread(&count, sizeof(count), 1, fp) < 1) {
		fprintf(stderr, "   Error reading from %s\n", file);
		return ERROR;
	}
	count = ntohl(count);

	if (count != 0) {
		table = (OverrideTableEntry *)
			calloc(count, sizeof(OverrideTableEntry));
		if (table == NULL) {
			fprintf(stderr, "   Unable to allocate memory "
					"for override table\n");
			return ERROR;
		}

		for (i=0; i < count; i++) {
			p = table[i].port;
			if ((fread(&n, sizeof(n), 1, fp) < 1) ||
			    (fread(&b, sizeof(b), 1, fp) < 1) ||
			    (fread(p, sizeof(*p), 1, fp) < 1)) {
				fprintf(stderr, "   Error reading from %s\n",
					file);
				free(table);
				return ERROR;
			}
			table[i].network.s_addr = ntohl(n.s_addr);
			table[i].bits = b;
			p[0].begin = ntohs(p[0].begin);
			p[0].end = ntohs(p[0].end);
			for (j=1, k=p[0].begin; (j<=k) && (j<LIST_SIZE); j++) {
				if (fread(&(p[j]), sizeof(*p), 1, fp) < 1) {
					fprintf(stderr, "   Error reading "
						"from %s\n", file);
				}
				p[j].begin = ntohs(p[j].begin);
				p[j].end = ntohs(p[j].end);
			}
		}

		req.ptr = table;
		req.bytes = count * sizeof(OverrideTableEntry);
		if (ioctl(theSocket, DIOCSOVERRIDE, &req)) {
			dfm_perror("DIOCSOVERRIDE");
			free(table);
			return ERROR;
		}
		free(table);
	}

	if (count > 0)
		fprintf(theOutput, "   Override table loaded: "
				   "%d entries\n", count);
	return NOERROR;
}

/*------------------------------------------------------------*/

static int
loadTCPresetTable(char *file, FILE *fp, u_int offset)
{
	/* Seek to the TCP reset table */
	fseek(fp, offset, SEEK_SET);

	/* Load the TCP reset tables */
	if (loadClassTable(fp, T_RST_IN, 0, "tcp_rst_in") == ERROR)
		return ERROR;
	if (loadClassTable(fp, T_RST_OUT, 0, "tcp_rst_out") == ERROR)
		return ERROR;

	fprintf(theOutput, "   TCP reset table loaded\n");
	return NOERROR;
}

/*------------------------------------------------------------*/

static int
loadFlags(char *file, FILE *fp, u_int offset)
{
	flagsReq flags;
	unsigned int i, count;
	u_int8_t flag, val;

	fseek(fp, offset, SEEK_SET);
	if (fread(&count, sizeof(count), 1, fp) < 1) {
		fprintf(stderr, "   Error reading from %s\n", file);
		return ERROR;
	}
	count = ntohl(count);

	if (ioctl(theSocket, DIOCGFLAGS, &flags)) {
		dfm_perror("DIOCSFLAGS");
		return ERROR;
	}

	for (i=0; i<count; i++) {
		if ((fread(&flag, sizeof(flag), 1, fp) < 1) ||
		    (fread(&val, sizeof(val), 1, fp) < 1)) {
			fprintf(stderr, "   Error reading from %s\n", file);
			return ERROR;
		} 
		switch (flag) {
		case FLAG_MULTICAST:
			flags.discardMulticast = val;
			break;
		case FLAG_NONIP:
			flags.discardNonIp = val;
			break;
		case FLAG_OTHERIP:
			flags.discardOtherIp = val;
			break;
		case FLAG_SUSPECTOFFSET:
			flags.discardSuspectOffset = val;
			break;
		case FLAG_FRAGMENTEDICMP:
			flags.discardFragmentedICMP = val;
			break;
		case FLAG_ATTACKICMP:
			flags.discardAttackICMP = val;
			break;
		}
	}

	if (ioctl(theSocket, DIOCSFLAGS, &flags)) {
		dfm_perror("DIOCSFLAGS");
		return ERROR;
	}

	if (count > 0)
		fprintf(theOutput, "   Flags loaded\n");
	return NOERROR;
}

/*------------------------------------------------------------*/

static int
loadTables(FILE *fp, fileHeader header, char *currTok)
{
	manageTablesReq mreq;

	/* INIT will allocate new tables for us to load data in */
	mreq.command = INIT_NEW;
	/* get the number of entries in the class table */
	fseek(fp, header.classes_offset, SEEK_SET);
	if (fread(&mreq.val, sizeof(mreq.val), 1, fp) != 1)
		return ERROR;
	mreq.val = ntohs(mreq.val);
	if (ioctl(theSocket, DIOCMNETWORK, &mreq)) {
		dfm_perror("DIOCMNETWORK");
		return ERROR;
	}

	if (loadAddrTable(currTok, fp, header.networks_offset)==ERROR)
		return ERROR;
	if (loadClassTables(currTok, fp, header.classes_offset)==ERROR)
		return ERROR;
	if (loadTCPresetTable(currTok, fp, header.tcpreset_offset)==ERROR)
		return ERROR;
	if (loadAccRejTable(currTok, fp, header.accept_offset, T_ACCTBL)==ERROR)
		return ERROR;
	if (loadAccRejTable(currTok, fp, header.reject_offset, T_REJTBL)==ERROR)
		return ERROR;
	if (loadOverrideTable(currTok, fp, header.override_offset)==ERROR)
		return ERROR;
	if (loadFlags(currTok, fp, header.flags_offset)==ERROR)
		return ERROR;

	/* ACTIVATE will make the new tables active and free the old tables */
	mreq.command = ACTIVATE_NEW;
	if (ioctl(theSocket, DIOCMNETWORK, &mreq)) {
		dfm_perror("DIOCMNETWORK");
		return ERROR;
	}

	return NO_ERROR;
}

/*------------------------------------------------------------*/

void
handleLoad()
{
	char *currTok;
	FILE *fp;
	fileHeader header;

	/*
	 * Make sure that we have all the parameters
	 */
	currTok = strtok(NULL, DELIMITERS);
	if ((findtoken(currTok) != T_FILTERS) ||
	    !(currTok = strtok(NULL, DELIMITERS)) ) {
		usageLoad(TERSE);
		return;
	}
	if (currTok && (*currTok == '?')) {
		usageLoad(VERBOSE);
		return;
	}

	/*
	 * Open the specified file
	 */
	if (!(fp = fopen(currTok, "r"))) {
		fprintf(theErrors, "   Can't open file '%s'\n", currTok);
		return;
	}

	/*
	 * Read the header and make sure it's correct
	 */
	if (fread(&header, sizeof(header), 1, fp) < 1) {
		fprintf(theErrors, "   Error reading from file '%s'\n",
			currTok);
		fclose(fp);
		return;
	}
	if (strncmp(header.magic, "DBFC", 4)) {
		fprintf(theErrors, "   File '%s' is invalid\n", currTok);
		fclose(fp);
		return;
	}
	if (strncmp(header.version, VERSION, 32)) {
		fprintf(theErrors, "   Config file '%s' was compiled by "
			"the wrong version of the compiler\n", currTok);
		fclose(fp);
		return;
	}
	/*
	 * Convert table offsets to host byte order
	 */
	header.networks_offset = ntohl(header.networks_offset);
	header.classes_offset = ntohl(header.classes_offset);
	header.accept_offset = ntohl(header.accept_offset);
	header.reject_offset = ntohl(header.reject_offset);
	header.override_offset = ntohl(header.override_offset);
	header.tcpreset_offset = ntohl(header.tcpreset_offset);
	header.flags_offset = ntohl(header.flags_offset);

	/*
	 * Load all the tables from the filter table file
	 */
	if (loadTables(fp, header, currTok)) {
		fclose(fp);
		return;
	}

	fclose(fp);
}

/*------------------------------------------------------------*/
