/*
 * Copyright (c) 1993,1994,1995,1997,1998
 *      Texas A&M University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Texas A&M University
 *      and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Developers:
 *     Russell Neeper, David K. Hess, Douglas Lee Schales, David R. Safford
 */

#include "opt_drawbridge.h"

#ifdef DRAWBRIDGE

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/mbuf.h>
#include <sys/ttycom.h>
#include <sys/ioccom.h>
#include <sys/filio.h>
#include <sys/sockio.h>

#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/syslog.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <machine/stdarg.h>

#include <net/if.h>
#include <net/if_types.h>
#include <net/if_llc.h>

#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/if_ether.h>
#include <netinet/if_fddi.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <machine/in_cksum.h>

/* #define DBDEBUG 1 */
#include <net/drawbridge.h>


int initialized = NO;

/* bit masks used for port mirroring */
int mirror_i1, mirror_i2, mirror_o1, mirror_o2, mirror_b;

Statistics DBstats = { NO }; 			/* default running state */

FilterConfig DBcfg = { DEFAULT_SYSL_FACILITY,	/* default log facility  */
		       DEFAULT_SYSL_MASK };	/* default log mask      */

/* MAC address hash table for bridge code */
/* Addresses are stored in network byte order */
HashEntry hashTable [MAX_BRIDGE_ADDRESSES];

pointerStruct active, new;
int ptrs_initialized = 0;


static void syslogMessage __P((u_int, ...));
static void freeAllPointers __P((pointerStruct *));
static int initTables __P((pointerStruct *, u_int16_t));
static void freeTree __P((addrTreeNode *, u_int));
static int addToTree __P((addrTreeNode *, in_addr_t, u_int, u_int16_t));
static int copyClass __P((int, portListReq *, PortListEntry *));
static int portListRequest __P((int, portListReq *, pointerStruct *));
static int addrTreeRequest __P((int, addrTreeReq *));
static int rejaccTableRequest __P((int,int, rejaccTableReq *, pointerStruct *));
static int overrideTableRequest __P((int, overrideTableReq *, pointerStruct *));
static int manageTablesRequest __P((manageTablesReq *));
static int bridgeRequest __P((int, bridgeReq *));
static int treeLookup __P((addrTreeNode *, u_int32_t, u_int16_t));
static int checkIncomingTcp __P((u_int32_t, u_int32_t, u_int16_t, u_int16_t));
static int checkOutgoingTcp __P((u_int32_t, u_int32_t, u_int16_t, u_int16_t));
static int checkIncomingUdp __P((u_int32_t, u_int32_t, u_int16_t, u_int16_t));
static int checkOutgoingUdp __P((u_int32_t, u_int32_t, u_int16_t, u_int16_t));
static int checkIncomingIcmp __P((u_int32_t, u_int32_t, u_int8_t));
static int checkOutgoingIcmp __P((u_int32_t, u_int32_t, u_int8_t));
static int checkRejectTable __P((struct ip *, u_int32_t, u_int32_t));
static int checkIncomingPacket __P((u_int16_t, u_int8_t *, u_int, int));
static int checkAcceptTable __P((struct ip *, u_int32_t, u_int32_t));
static int checkOutgoingPacket __P((u_int16_t, u_int8_t *, u_int));
static int bridge __P((int, HardwareAddr *, HardwareAddr *));
static int db_open __P((dev_t, int, int, struct proc *));
static int db_close __P((dev_t, int, int, struct proc *));
static int db_ifpromisc __P((struct ifnet *, int));
static int db_gethwaddr __P((struct ifnet *, HardwareAddr *));
static int db_getIPaddr __P((struct ifnet *, in_addr_t *));
static int initRequest __P((initReq *));
static int db_ioctl __P((dev_t, u_long, caddr_t, int, struct proc *));
static void db_drvinit __P((void *));


#ifdef DEVFS
#include <sys/devfsext.h>
void	*devfs_token;
#endif

#define CDEV_MAJOR 20		/* 20 is for locally defined char devices */

#if defined(__FreeBSD__) && __FreeBSD__ >= 4

static struct cdevsw db_cdevsw = {
	db_open,      db_close,   noread, nowrite,
	db_ioctl,     nopoll,     nommap, nostrategy,
	"drawbridge", CDEV_MAJOR, nodump, nopsize,  0,   -1
};

#else

static int db_poll __P((dev_t, int, struct proc *));

static struct cdevsw db_cdevsw = {
	db_open,   db_close, noread,    nowrite,
	db_ioctl,  nostop,   nullreset, nodevtotty,
	db_poll,   nommap,   NULL,      "drawbridge",  NULL,  -1
};

static int
db_poll(dev_t dev, int events, struct proc *p)
{
        return (ENODEV);
}

#endif

/* ------------------------------------------------------------------------ */
/*
 * Thanks to Klaus-Peter Kossakowski and Uwe Ellermann at DFN-CERT
 * for contributing much of the syslog implementation.
 *
 * All parameters are in host byte ordering.
 */

/*
 * Note:  If this table is modified, the corresponding
 * enum struct in drawbridge.h must also be modified.
 */
static SyslogMessageEntry syslogMessages[] = {
	{ "unknown event",				LOG_WARNING },
	{ "cold start",					LOG_INFO    },
	{ "incoming class D:",				LOG_WARNING },
	{ "outgoing class D:",				LOG_WARNING },
	{ "incoming port:",				LOG_WARNING },
	{ "outgoing port:",				LOG_WARNING },
	{ "incoming type:",				LOG_WARNING },
	{ "outgoing type:",				LOG_WARNING },
	{ "incoming via reject table:",			LOG_WARNING },
	{ "outgoing via accept table:",			LOG_WARNING },
	{ "outgoing via override table:",		LOG_WARNING },
	{ "incoming header too short:",			LOG_WARNING },
	{ "outgoing header too short:",			LOG_WARNING },
	{ "incoming D-O-S attack:",			LOG_WARNING },
	{ "outgoing D-O-S attack:",			LOG_WARNING },
	{ "incoming IP:",				LOG_NOTICE  },
	{ "outgoing IP:",				LOG_WARNING },
	{ "incoming fragment with IP offset == 1:",	LOG_WARNING },
	{ "outgoing fragment with IP offset == 1:",	LOG_WARNING },
	{ "incoming fragment:",				LOG_WARNING },
	{ "outgoing fragment:",				LOG_WARNING },
	{ "incoming MAC layer protocol:",		LOG_INFO    },
	{ "outgoing MAC layer protocol:",		LOG_WARNING },
	{ NULL } /* Marks the end of the messages. */
};

static const char *syslogDosTypes[] = { "\"Smurf\"", "\"Pong\"" };

static void
syslogMessage(u_int eventNo,...)
{
        va_list ap;
	int protocolNo;
	in_addr_t srcAddr;
	in_addr_t dstAddr;
	int srcPort;
	int dstPort;
	int type;
	int protocol;

	/*
	 * If this message is disabled then just return.
	 */
	if ( !((1UL << eventNo) & DBcfg.logMask) )
		return;

	switch ( eventNo ) {
		case SYSL_IN_OFFSET:
		case SYSL_OUT_OFFSET:
		case SYSL_IN_FRAG:
		case SYSL_OUT_FRAG:
		case SYSL_IN_REJECT:
		case SYSL_OUT_ACCEPT:
		case SYSL_IN_PROT:
		case SYSL_OUT_PROT: {	/* IP - general */
			va_start(ap,eventNo);

			protocolNo = va_arg(ap,int);
			srcAddr    = va_arg(ap,in_addr_t);
			dstAddr    = va_arg(ap,in_addr_t);

			va_end(ap);

			log(	(DBcfg.logFacility |
				 syslogMessages[eventNo].priority),
				"drawbridge: %s protocol %d " \
				"from %u.%u.%u.%u to %u.%u.%u.%u\n",
				syslogMessages[eventNo].message, protocolNo,
				srcAddr.s_b[3], srcAddr.s_b[2],
				srcAddr.s_b[1], srcAddr.s_b[0],
				dstAddr.s_b[3], dstAddr.s_b[2],
				dstAddr.s_b[1], dstAddr.s_b[0]
			);
			break;
		}

		case SYSL_IN_CLASSD:
		case SYSL_OUT_CLASSD:
		case SYSL_IN_PORT:
		case SYSL_OUT_PORT:
		case SYSL_OUT_OVERRIDE:
		case SYSL_IN_LENGTH:
		case SYSL_OUT_LENGTH: {	/* IP - TCP/UDP */
			va_start(ap,eventNo);

			protocolNo = va_arg(ap,int);
			srcAddr    = va_arg(ap,in_addr_t);
			dstAddr    = va_arg(ap,in_addr_t);
			srcPort    = va_arg(ap,int);
			dstPort    = va_arg(ap,int);

			va_end(ap);

			log(	(DBcfg.logFacility |
				 syslogMessages[eventNo].priority),
				"drawbridge: %s %s from %u.%u.%u.%u " \
				"port %u to %u.%u.%u.%u port %u\n",
				(protocolNo == IPPROTO_UDP) ? "UDP" : "TCP",
				syslogMessages[eventNo].message,
				srcAddr.s_b[3], srcAddr.s_b[2],
				srcAddr.s_b[1], srcAddr.s_b[0],
				srcPort,
				dstAddr.s_b[3], dstAddr.s_b[2],
				dstAddr.s_b[1], dstAddr.s_b[0],
				dstPort
			);
			break;
		}

		case SYSL_IN_TYPE: 	/* IP - ICMP */
		case SYSL_OUT_TYPE: {
			va_start(ap,eventNo);

			srcAddr = va_arg(ap,in_addr_t);
			dstAddr = va_arg(ap,in_addr_t);
			type    = va_arg(ap,int);

			va_end(ap);

			log(	(DBcfg.logFacility |
				 syslogMessages[eventNo].priority),
				"drawbridge: ICMP %s from " \
				"%u.%u.%u.%u type %u to %u.%u.%u.%u\n",
				syslogMessages[eventNo].message,
				srcAddr.s_b[3], srcAddr.s_b[2],
				srcAddr.s_b[1], srcAddr.s_b[0],
				type,
				dstAddr.s_b[3], dstAddr.s_b[2],
				dstAddr.s_b[1], dstAddr.s_b[0]
			);
			break;
		}

		case SYSL_IN_DOS:	/* Denial of service attack */
		case SYSL_OUT_DOS: {
			va_start(ap,eventNo);

			srcAddr = va_arg(ap,in_addr_t);
			dstAddr = va_arg(ap,in_addr_t);
			type    = va_arg(ap,int);

			va_end(ap);

			log(	(DBcfg.logFacility |
				syslogMessages[eventNo].priority),
				"drawbridge: %s %s from " \
				"%u.%u.%u.%u to %u.%u.%u.%u\n",
				syslogDosTypes[type],
				syslogMessages[eventNo].message,
				srcAddr.s_b[3], srcAddr.s_b[2],
				srcAddr.s_b[1], srcAddr.s_b[0],
				dstAddr.s_b[3], dstAddr.s_b[2],
				dstAddr.s_b[1], dstAddr.s_b[0]
			);
			break;
		}

		case SYSL_IN_FILTER:
		case SYSL_OUT_FILTER: {	/* MAC layer protocol */
			va_start(ap,eventNo);
			protocol = va_arg(ap,int);
			va_end(ap);

			log(	(DBcfg.logFacility |
				 syslogMessages[eventNo].priority),
				"drawbridge: %s 0x%04x\n",
				syslogMessages[eventNo].message, protocol
			);
			break;
		}

		case SYSL_COLD_START:	/* Startup message */
			log(	(DBcfg.logFacility |
				 syslogMessages[eventNo].priority),
				"drawbridge: %s\n",
				syslogMessages[eventNo].message
			);
			break;

		default:		/* Should never get here */
			log(	(DBcfg.logFacility | LOG_WARNING),
				"drawbridge: unknown event\n"
			);
			break;
	}
}

/* ------------------------------------------------------------------------ */

static void
freeAllPointers(pointerStruct *p)
{
	if (p->addrTree)	freeTree(p->addrTree, 0);
	if (p->t_dst_in)	free(p->t_dst_in, M_DEVBUF);
	if (p->t_dst_out)	free(p->t_dst_out, M_DEVBUF);
        if (p->t_src_in)	free(p->t_src_in, M_DEVBUF);
        if (p->t_src_out)	free(p->t_src_out, M_DEVBUF);
        if (p->u_dst_in)	free(p->u_dst_in, M_DEVBUF);
        if (p->u_dst_out)	free(p->u_dst_out, M_DEVBUF);
        if (p->u_src_in)	free(p->u_src_in, M_DEVBUF);
        if (p->u_src_out)	free(p->u_src_out, M_DEVBUF);
        if (p->i_typ_in)	free(p->i_typ_in, M_DEVBUF);
        if (p->i_typ_out)	free(p->i_typ_out, M_DEVBUF);
        if (p->t_rst_in)	free(p->t_rst_in, M_DEVBUF);
        if (p->t_rst_out)	free(p->t_rst_out, M_DEVBUF);
        if (p->acceptTable)	free(p->acceptTable, M_DEVBUF);
        if (p->acceptTree)	freeTree(p->acceptTree, 0);
        if (p->rejectTable)	free(p->rejectTable, M_DEVBUF);
        if (p->rejectTree)	freeTree(p->rejectTree, 0);
        if (p->overrideTable)	free(p->overrideTable, M_DEVBUF);
        if (p->overrideTree)	freeTree(p->overrideTree, 0);

	bzero(p, sizeof(pointerStruct));
}

/* ------------------------------------------------------------------------ */

static int
initTables(pointerStruct *p, u_int16_t classes)
{
	/* Must have at least one class for default */
	if (classes == 0)
		classes = 1;

	/* Free all allocated memory referenced by pointer struct */
	freeAllPointers(p);

	/* Build a new default config */
	p->addrTree  = MALLOC_NODE;
	p->t_dst_in  = MALLOC_CLASS(classes);
	p->t_dst_out = MALLOC_CLASS(classes);
	p->t_src_in  = MALLOC_CLASS(classes);
	p->t_src_out = MALLOC_CLASS(classes);
	p->u_dst_in  = MALLOC_CLASS(classes);
	p->u_dst_out = MALLOC_CLASS(classes);
	p->u_src_in  = MALLOC_CLASS(classes);
	p->u_src_out = MALLOC_CLASS(classes);
	p->i_typ_in  = MALLOC_CLASS(classes);
	p->i_typ_out = MALLOC_CLASS(classes);
	p->t_rst_in  = MALLOC_CLASS(1);
	p->t_rst_out = MALLOC_CLASS(1);
	if ((p->addrTree == NULL) ||
	    (p->t_dst_in  == NULL) || (p->t_dst_out == NULL) ||
	    (p->t_src_in  == NULL) || (p->t_src_out == NULL) ||
	    (p->u_dst_in  == NULL) || (p->u_dst_out == NULL) ||
	    (p->u_src_in  == NULL) || (p->u_src_out == NULL) ||
	    (p->i_typ_in  == NULL) || (p->i_typ_out == NULL) ||
	    (p->t_rst_in  == NULL) || (p->t_rst_out == NULL)) {
		freeAllPointers(p);
		return ERROR_NO_MEMORY;
	}
	p->num_classes = classes;
	bzero(p->addrTree,  sizeof(addrTreeNode));
	bzero(p->t_dst_in,  classes * LIST_TABLE_SIZE);
	bzero(p->t_dst_out, classes * LIST_TABLE_SIZE);
	bzero(p->t_src_in,  classes * LIST_TABLE_SIZE);
	bzero(p->t_src_out, classes * LIST_TABLE_SIZE);
	bzero(p->u_dst_in,  classes * LIST_TABLE_SIZE);
	bzero(p->u_dst_out, classes * LIST_TABLE_SIZE);
	bzero(p->u_src_in,  classes * LIST_TABLE_SIZE);
	bzero(p->u_src_out, classes * LIST_TABLE_SIZE);
	bzero(p->i_typ_in,  classes * LIST_TABLE_SIZE);
	bzero(p->i_typ_out, classes * LIST_TABLE_SIZE);
	bzero(p->t_rst_in,  LIST_TABLE_SIZE);
	bzero(p->t_rst_out, LIST_TABLE_SIZE);

	/* Fill in the class table defaults */
        p->t_dst_in[0].begin = 2;	/* TCP - two entries follow */
        p->t_dst_in[1].begin = 25;	/* TCP - Mail in            */
        p->t_dst_in[1].end = 25;
        p->t_dst_in[2].begin = 53;	/* TCP - Name service in */
        p->t_dst_in[2].end = 53;

        p->t_dst_out[0].begin = 1;	/* TCP - one entry follows */
        p->t_dst_out[1].begin = 0;	/* TCP - Everything out    */
        p->t_dst_out[1].end = 0xFFFF;

        p->t_src_in[0].begin = 1;	/* TCP - one entry follows       */
        p->t_src_in[1].begin = 20;	/* TCP - FTP data connections in */
        p->t_src_in[1].end = 20;

        p->u_dst_in[0].begin = 3;	/* UDP - three entries follow       */
        p->u_dst_in[1].begin = 0;	/* UDP - Block TFTP & Portmapper in */
        p->u_dst_in[1].end = 68;
        p->u_dst_in[2].begin = 70;
        p->u_dst_in[2].end = 110;
        p->u_dst_in[3].begin = 112;
        p->u_dst_in[3].end = 0xFFFF;

	p->u_dst_out[0].begin = 1;	/* UDP - one entry follows */
	p->u_dst_out[1].begin = 0;	/* UDP - Everything out    */
	p->u_dst_out[1].end = 0xFFFF;

	p->i_typ_in[0].begin = 1;	/* ICMP - one entry follows     */
	p->i_typ_in[1].begin = 0;	/* ICMP - Everything in         */
	p->i_typ_in[1].end = 0xFF;

	p->i_typ_out[0].begin = 1;	/* ICMP - one entry follows     */
	p->i_typ_out[1].begin = 0;	/* ICMP - Everything out        */
	p->i_typ_out[1].end = 0xFF;

	return NO_ERROR;
}

/*-----------------------------------------------------------------------*/

static void
freeTree(addrTreeNode *tbl, u_int level)
{
        int i;

        if (!tbl || (level > 3)) {
		printf("freeTree called with invalid parameters! "
			"(tbl:%p, level:%d)\n", tbl, level);
                return;
	}

	if (level == 3) {
		free((addrTreeLeaf *)tbl, M_DEVBUF);
		return;
	}

        for (i=0; i<256; i++) {
                if (!tbl->flg[i] && tbl->u[i].p) {
			if (level == 2)
				free(tbl->u[i].pd, M_DEVBUF);
			else
				freeTree(tbl->u[i].p, level + 1);
		}
	}

	free(tbl, M_DEVBUF);
}

/* ------------------------------------------------------------------------ */
/*
 * addr is expected to be in host (reverse) byte order
 * This function only works correctly on little endian CPUs
 */

static int
addToTree(addrTreeNode *tbl, in_addr_t addr, u_int bits, u_int16_t class)
{
	in_addr_t mask;
	u_int32_t i, byte, start, end, n, level;
	addrTreePtr ptr;

        if (!tbl || (bits > 32))
                return -1;

	/* Find the rightmost non zero byte in the mask */
	mask.s_addr = 0xFFFFFFFFUL << ((bits < 32) ? 32 - bits : 0);
	for (byte=0; (byte < 3) && (mask.s_b[byte] == 0); byte++);

	/* Determine the range of addresses to be added to the tree */
	start = addr.s_b[byte] & mask.s_b[byte];
	end = start + (u_int8_t) ~mask.s_b[byte];

	for (i=start; i<=end; i++) {

		ptr = (addrTreePtr)tbl;
		addr.s_b[byte] = i;

		/* 'byte' specifies how deep in the tree to go */
		/* 'level' keeps track of how deep we are: 0, 1, 2, or 3 */
		for (n=byte, level=0; n < 4; n++, level++) {

			/* Use the next byte of the ip address as an index */
			/* into the array of pointers/classes of this node. */
			register u_int8_t b = addr.s_b[3 - level];

			/* If n equal 3, then this is the last byte of */
			/* the ip address to be put in the tree.  */
			if (n == 3) {
				if (level == 3) {
					/* Level 3 contains only classes */
					ptr.pd->c[b] = class;

				} else {
					/* Levels 0-2 can contain either a   */
					/* class or a pointer.  If a pointer */
					/* is stored here, then prune the    */
					/* tree from here up.                */
					if (!ptr.p->flg[b] && ptr.p->u[b].p)
						freeTree(ptr.p->u[b].p,level+1);
					ptr.p->u[b].p = NULL;

					/* Now save the class in this slot.  */
					ptr.p->u[b].c = class;
					ptr.p->flg[b] = 1;
				}

			/* If this is a pointer... */
			} else if (!ptr.p->flg[b] && ptr.p->u[b].p) {

				/* then move up the tree one level */
				ptr.p = ptr.p->u[b].p;

			/* Still more to go but we ran into a class   */
			/* index.  Add another level and move the     */
			/* class index to each slot in the new level. */
			} else {
				register u_int j;
				addrTreePtr next;

				/* If we are at level 2, then the */
				/*  next level is different.      */
				if (level == 2) {
					u_int tmp_c = ptr.p->u[b].c;

					if (!(next.pd = MALLOC_LEAF))
						return 1;
					for (j=0; j<256; j++) {
						next.pd->c[j] = tmp_c;
					}
				} else {
					void *tmp_p = ptr.p->u[b].p;
					u_int tmp_f = ptr.p->flg[b];

					if (!(next.p = MALLOC_NODE))
						return 1;
					for (j=0; j<256; j++) {
						next.p->u[j].p = tmp_p;
						next.p->flg[j] = tmp_f;
					}
				}
				/* Replace the class with the new pointer,  */
				/* set the flag to 0 to indicate a pointer, */
				/* and then move up the tree one level.     */ 
				ptr.p->u[b].p = next.p;
				ptr.p->flg[b] = 0;
				ptr.p = next.p;
			}
		}
	}
	return 0;
}

/* ------------------------------------------------------------------------ */

static int
copyClass(int rw, portListReq *u, PortListEntry *kptr)
{
	if (kptr == NULL)
		return ERROR_NET_NOT_FOUND; /* XXX */
	if (rw == WRITE)
		return copyin(u->ptr, kptr, u->bytes);
	if (rw == READ)
		return copyout(kptr, u->ptr, u->bytes);
	return ERROR_INVALID_OP;
}

/* ------------------------------------------------------------------------ */

static int
portListRequest(int rw, portListReq *u, pointerStruct *p)
{
	if (u->class + ((u->bytes / LIST_TABLE_SIZE) +
		        (u->bytes % LIST_TABLE_SIZE) ? 1 : 0) > p->num_classes)
		return ERROR_OUT_OF_RANGE;

	switch (u->id) {
	case T_DST_IN:
		return copyClass(rw, u ,&(p->t_dst_in[u->class * LIST_SIZE]));
	case T_DST_OUT:
		return copyClass(rw, u ,&(p->t_dst_out[u->class * LIST_SIZE]));
	case T_SRC_IN:
		return copyClass(rw, u ,&(p->t_src_in[u->class * LIST_SIZE]));
	case T_SRC_OUT:
		return copyClass(rw, u ,&(p->t_src_out[u->class * LIST_SIZE]));
	case U_DST_IN:
		return copyClass(rw, u ,&(p->u_dst_in[u->class * LIST_SIZE]));
	case U_DST_OUT:
		return copyClass(rw, u ,&(p->u_dst_out[u->class * LIST_SIZE]));
	case U_SRC_IN:
		return copyClass(rw, u ,&(p->u_src_in[u->class * LIST_SIZE]));
	case U_SRC_OUT:
		return copyClass(rw, u ,&(p->u_src_out[u->class * LIST_SIZE]));
	case I_TYP_IN:
		return copyClass(rw, u ,&(p->i_typ_in[u->class * LIST_SIZE]));
	case I_TYP_OUT:
		return copyClass(rw, u ,&(p->i_typ_out[u->class * LIST_SIZE]));
	case T_RST_IN:
		if (u->class > 0)
			return ERROR_OUT_OF_RANGE;
		return copyClass(rw, u, p->t_rst_in);
	case T_RST_OUT:
		if (u->class > 0)
			return ERROR_OUT_OF_RANGE;
		return copyClass(rw, u, p->t_rst_out);
	default:
		return ERROR_UNKNOWN_CLASS;
	}
}

/* ------------------------------------------------------------------------ */

static int
addrTreeRequest(int rw, addrTreeReq *u)
{
	register u_int i;

	switch (rw) {
        case READ:
		/* Lookup the specified IP cidr(s) in the 'active' config */
		for (i=0; (i < 128) && (i < u->num); i++)
			u->class[i] = treeLookup(active.addrTree,
					u->addr[i].s_addr, active.num_classes);
		break;
	
        case WRITE:
		if (new.addrTree == NULL) 
			return ERROR_NET_NOT_FOUND; /* XXX */

		/* Add the specified IP cidr(s) to the 'new' config */
		for (i=0; (i < u->num) && (i < 128); i++) {
			if (addToTree(new.addrTree, u->addr[i],
				     u->bits[i], u->class[i])) {
				freeTree(new.addrTree, 0);
				new.addrTree = NULL;
				return ERROR_NO_MEMORY;
			}
		}
		break;

	default:
		return ERROR_INVALID_OP;
	}

	return NO_ERROR;
}

/* ------------------------------------------------------------------------ */

static int
rejaccTableRequest(int rw, int id, rejaccTableReq *u, pointerStruct *p)
{
	register u_int i;
	RejAccTableEntry *table;
	addrTreeNode *tree;

	if ((id != T_REJTBL) && (id != T_ACCTBL))
		return ERROR_INVALID_OP;

	switch (rw) {
        case READ:
		u->bytes = (id == T_REJTBL) ? p->reject_bytes : p->accept_bytes;
		u->flag = treeLookup((id == T_REJTBL) ? active.rejectTree :
					active.acceptTree, u->addr, 2);
		table = (id == T_REJTBL) ? p->rejectTable : p->acceptTable;
		if ((u->bytes != 0) && (table == NULL))
			return ERROR_NET_NOT_FOUND; /* XXX */
		/* If it exists and the user requested it, */
		/* copy the table up to user space.        */
		if ((u->ptr != NULL) && (table != NULL)) {
			if ((i = copyout(table, u->ptr, u->bytes)) != 0)
				return i;
		}
		break;
	
        case WRITE:
		/* Perform a few sanity checks */
		if ((u->bytes == 0) || (u->bytes % sizeof(RejAccTableEntry)))
			return ERROR_INVALID_OP;
		if (id == T_REJTBL) {
			if ((p->rejectTable != NULL) || (p->rejectTree != NULL))
				return ERROR_NET_NOT_FOUND; /* XXX */
		} else {
			if ((p->acceptTable != NULL) || (p->acceptTree != NULL))
				return ERROR_NET_NOT_FOUND; /* XXX */
		}

		/*
		 * Allocate the memory for the table and tree.  The tree
		 * is used for packet filtering.  The table is only used
		 * to keep track of the networks/hosts to show to the user.
		 */
		table = (RejAccTableEntry *)malloc(u->bytes,M_DEVBUF,M_NOWAIT);
		if (table == NULL)
			return ERROR_NO_MEMORY;
		tree = MALLOC_NODE;
		if (tree == NULL) {
			free(table, M_DEVBUF);
			return ERROR_NO_MEMORY;
		}
		bzero(tree, sizeof(addrTreeNode));
		bzero(table, u->bytes);

		/* Copy the table in from user space */
		if ((i = copyin(u->ptr, table, u->bytes)) != 0)
			return i;
		/*
		 * Populate the tree from the table - keep the table around
		 * to spit back out when requested by the user interface.
		 */
		for (i=u->bytes/sizeof(RejAccTableEntry); i; i--) {
			register int r;

			if ((r=addToTree(tree, table[i-1].network,
				      table[i-1].bits, table[i-1].flag))) {
				freeTree(tree, 0);
				free(table, M_DEVBUF);
				return (r>0) ? ERROR_NO_MEMORY :
					ERROR_INVALID_OP;
			}
		}
		if (id == T_REJTBL) {
			p->rejectTable = table;
			p->rejectTree = tree;
			p->reject_bytes = u->bytes;
		} else {
			p->acceptTable = table;
			p->acceptTree = tree;
			p->accept_bytes = u->bytes;
		}
		break;

	default:
		return ERROR_INVALID_OP;
	}

	return NO_ERROR;
}

/* ------------------------------------------------------------------------ */

static int
overrideTableRequest(int rw, overrideTableReq *u, pointerStruct *p)
{
	u_int i,j,r;
	OverrideTableEntry *table;
	addrTreeNode *tree;

	switch (rw) {
        case READ:
		u->bytes = p->override_bytes;
		u->index = treeLookup(active.overrideTree,
				      u->addr, active.num_override);
		table = p->overrideTable;
		if ((u->bytes != 0) && (table == NULL))
			return ERROR_NET_NOT_FOUND; /* XXX */
		/* If it exists and the user requested it, */
		/* copy the table up to user space.        */
		if ((u->ptr != NULL) && (table != NULL)) {
			if ((i = copyout(table, u->ptr, u->bytes)) != 0)
				return i;
		}
		break;
	
        case WRITE:
		/* Perform a few sanity checks */
		if ((u->bytes == 0) || (u->bytes % sizeof(OverrideTableEntry)))
			return ERROR_INVALID_OP;
		if ((p->overrideTable != NULL) || (p->overrideTree != NULL))
			return ERROR_NET_NOT_FOUND; /* XXX */

		/* Allocate the memory for the table and tree. */
		table = (OverrideTableEntry *)
				malloc(u->bytes, M_DEVBUF, M_NOWAIT);
		if (table == NULL)
			return ERROR_NO_MEMORY;
		tree = MALLOC_NODE;
		if (tree == NULL) {
			free(table, M_DEVBUF);
			return ERROR_NO_MEMORY;
		}
		bzero(tree, sizeof(addrTreeNode));
		bzero(table, u->bytes);

		/* Copy the table in from user space */
		if ((i = copyin(u->ptr, table, u->bytes)) != 0)
			return i;

		/* Populate the tree from the table */
		for (i=u->bytes/sizeof(OverrideTableEntry),j=1; i; i--,j++) {
			if ((r=addToTree(tree, table[i-1].network,
				      table[i-1].bits, j))) {
				freeTree(tree, 0);
				free(table, M_DEVBUF);
				return (r>0) ? ERROR_NO_MEMORY :
					ERROR_INVALID_OP;
			}
		}
		p->overrideTable = table;
		p->overrideTree = tree;
		p->override_bytes = u->bytes;
		p->num_override = (u->bytes / sizeof(OverrideTableEntry)) + 1;
		break;

	default:
		return ERROR_INVALID_OP;
	}

	return NO_ERROR;
}

/* ------------------------------------------------------------------------ */

static int
manageTablesRequest(manageTablesReq *u)
{
	pointerStruct tmp;
	int rcode;
	u_int s;

	switch (u->command) {

	/* Move the 'new' config to 'active' and deallocate the old config */
	case ACTIVATE_NEW:
		if ((new.num_classes == 0) || (new.addrTree == NULL) ||
		    (new.t_dst_in == NULL) || (new.t_dst_out == NULL) ||
		    (new.t_src_in == NULL) || (new.t_src_out == NULL) ||
		    (new.u_dst_in == NULL) || (new.u_dst_out == NULL) ||
		    (new.u_src_in == NULL) || (new.u_src_out == NULL) ||
		    (new.i_typ_in == NULL) || (new.i_typ_out == NULL) ||
		    (new.t_rst_in == NULL) || (new.t_rst_out == NULL))
			return ERROR_NET_NOT_FOUND; /* XXX */
		memcpy(&tmp, &active, sizeof(pointerStruct));
		/* The swap is done all at once between network interrupts. */
		s = splimp();
		memcpy(&active, &new, sizeof(pointerStruct));
		splx(s);
		bzero(&new, sizeof(pointerStruct));
		freeAllPointers(&tmp);
		break;

	/* Remove all custom filters and revert to defaults */
	case CLEAR_ALL:
		/* (Re)initialize the 'new' config. */
		rcode = initTables(&new, 1);
		if (rcode)
			return rcode;
		memcpy(&tmp, &active, sizeof(pointerStruct));
		/* Move the 'new' config to the 'active' config */
		s = splimp();
		memcpy(&active, &new, sizeof(pointerStruct));
		splx(s);
		bzero(&new, sizeof(pointerStruct));
		/* Deallocate the old config */
		freeAllPointers(&tmp);
		break;

	/* (Re)initialize the 'new' config.  Needs to   */
	/* be done before loading a new set of filters. */
	case INIT_NEW:
		rcode = initTables(&new, u->val);
		if (rcode)
			return rcode;
		break;

	}

	return NO_ERROR;
}

/* ------------------------------------------------------------------------ */

static int
bridgeRequest(int rw, bridgeReq *r)
{
	u_int s;

	if (rw != READ)
		return ERROR_INVALID_OP;

	if (r->index >= MAX_BRIDGE_ADDRESSES)
		return ERROR_OUT_OF_RANGE;

	/* Block network interrupts so the table entry */
	/* doesn't change while we are copying it. */
	s = splimp();
	bcopy(&hashTable[r->index], &r->t, sizeof(r->t));
	splx(s);

	return NO_ERROR;
}

/*-----------------------------------------------------------------------*/
/*
 * 'addr' is expected to be in host (reverse) byte order.
 * This function only works correctly on little endian CPUs.
 */

static int
treeLookup(addrTreeNode *tbl, u_int32_t addr, u_int16_t max_class)
{
        register addrTreePtr t1,t2;
	register in_addr_t ip;

	if (tbl == NULL)
		return DEFAULT_ACCESS;
	ip = (in_addr_t)addr;

        /*
         * lookup first octet of address - flg=1: class flg=0: pointer
         */
        if ((t1.p = tbl->u[ip.s_b[3]].p) == NULL)
                return DEFAULT_ACCESS;
        if (tbl->flg[ip.s_b[3]])
                return ((t1.c < max_class) ? t1.c : DEFAULT_ACCESS);

        /*
         * lookup second octet of address - flg=1: class flg=0: pointer
         * t1.p = tbl->u[ip.s_b[3]].p
         */
        if ((t2.p = t1.p->u[ip.s_b[2]].p) == NULL)
                return DEFAULT_ACCESS;
        if (t1.p->flg[ip.s_b[2]])
                return ((t2.c < max_class) ? t2.c : DEFAULT_ACCESS);

        /*
         * lookup third octet of address - flg=1: class flg=0: pointer
         * t2.p = tbl->u[ip.s_b[3]].p->u[ip.s_b[2]].p
         */
        if ((t1.pd = t2.p->u[ip.s_b[1]].pd) == NULL)
                return DEFAULT_ACCESS;
        if (t2.p->flg[ip.s_b[1]])
                return ((t1.c < max_class) ? t1.c : DEFAULT_ACCESS);

        /*
         * lookup fourth octet of address - always a class
         * t1.pd = tbl->u[ip.s_b[3]].p->u[ip.s_b[2]].p->u[ip.s_b[1]].pd
         */
        t2.c = t1.pd->c[ip.s_b[0]];
        return ((t2.c < max_class) ? t2.c : DEFAULT_ACCESS);
}

/* ------------------------------------------------------------------------ */
/*
 * Parameters are expected to be in host byte order
 */

static int
checkIncomingTcp(u_int32_t srcAddr, u_int32_t dstAddr,
		 u_int16_t srcPort, u_int16_t dstPort)
{
	register int i;
	int class;
	PortListEntry *portList;

	/*
	 * Check for multicast
	 */
	if (IN_CLASSD(dstAddr)) {
		if (DBcfg.discardMulticast) {
			DBstats.f_outsideTcpMcast++;
			syslogMessage(SYSL_IN_CLASSD, IPPROTO_TCP,
			      	srcAddr, dstAddr, (int)srcPort, (int)dstPort);
			return DISCARD_PKT;
		}
		return ALLOW_PKT;
	}

	/*
	 * Do the lookup to get the class.
	 */
	class = treeLookup(active.addrTree, dstAddr, active.num_classes);
	portList = &active.t_dst_in[class * LIST_SIZE];

	/*
	 * See if the destination port is allowed by the 'in' port list
	 */
	for (i=1; i <= portList[0].begin; i++) {
		if ((dstPort >= portList[i].begin) &&
		    (dstPort <= portList[i].end))
	    		return ALLOW_PKT;
	}

	/*
	 * If the destination port is not allowed but is over 900, then
	 * check the source port.  NOTE: This is 900 instead of 1024 because
	 * some broken FTP implementations use ports below 1024.  If you want
	 * to enforce the use of non-privileged ports, change this to >= 1024.
	 */
	if (dstPort > 900) {
		portList = &active.t_src_in[class * LIST_SIZE];

		for (i=1; i <= portList[0].begin; i++) {
			if ((srcPort >= portList[i].begin) &&
			    (srcPort <= portList[i].end))
				return ALLOW_PKT;
		}
	}

	DBstats.f_outsideTcpPort++;
	syslogMessage(SYSL_IN_PORT, IPPROTO_TCP,
		      srcAddr, dstAddr, (int)srcPort, (int)dstPort);

	/* Check the TCP reset table */
	for (i=1; i <= active.t_rst_in[0].begin; i++) {
		if ((dstPort >= active.t_rst_in[i].begin) &&
		    (dstPort <= active.t_rst_in[i].end)) {
			DBstats.outsideTCPreset++;
			return DISCARD_AND_ANSWER_PKT;
		}
	}
	return DISCARD_PKT;
}

/* ------------------------------------------------------------------------ */
/*
 * Parameters are expected to be in host byte order
 */

static int
checkOutgoingTcp(u_int32_t srcAddr, u_int32_t dstAddr,
		 u_int16_t srcPort, u_int16_t dstPort)
{
	register int i, j;
	int class;
	PortListEntry *portList;

	/*
	 * Check for multicast
	 */
	if (IN_CLASSD(dstAddr)) {
		if (DBcfg.discardMulticast) {
			DBstats.f_insideTcpMcast++;
			syslogMessage(SYSL_OUT_CLASSD,IPPROTO_TCP,
				srcAddr, dstAddr, (int)srcPort, (int)dstPort);
			return DISCARD_PKT;
		}
		return ALLOW_PKT;
	}

	/*
	 * Do the lookup to get the class.
	 */
	class = treeLookup(active.addrTree, srcAddr, active.num_classes);
	portList = &active.t_dst_out[class * LIST_SIZE];

	/*
	 * See if the destination port is allowed by the 'out' port list
	 */
	for (i=1; i <= portList[0].begin; i++) {
		if ((dstPort >= portList[i].begin) &&
		    (dstPort <= portList[i].end))
			return ALLOW_PKT;
	}

	/*
	 * If the destination port is not allowed but is over 900, then
	 * check the source port.  NOTE: This is 900 instead of 1024 because
	 * some broken FTP implementations use ports below 1024.  If you want
	 * to enforce the use of non-privileged ports, change this to >= 1024.
	 */
	if (dstPort > 900) {
		portList = &active.t_src_out[class * LIST_SIZE];

		for (i=1; i <= portList[0].begin; i++) {
			if ((srcPort >= portList[i].begin) &&
			    (srcPort <= portList[i].end))
				return ALLOW_PKT;
		}
	}

	/*
	 * Still not allowed so check the outgoing packet to see
	 * if the destination is in the override table allow list.
	 */
	if ((i = treeLookup(active.overrideTree,
			    dstAddr, active.num_override)) == 0) {
		DBstats.f_insideTcpPort++;
		syslogMessage(SYSL_OUT_PORT, IPPROTO_TCP,
			srcAddr, dstAddr, (int)srcPort, (int)dstPort);
		/* Check the TCP reset table */
		for (i=1; i <= active.t_rst_out[0].begin; i++) {
			if ((dstPort >= active.t_rst_out[i].begin) &&
			    (dstPort <= active.t_rst_out[i].end)) {
				DBstats.insideTCPreset++;
				return DISCARD_AND_ANSWER_PKT;
			}
		}
    		/* Dest addr was not in the override table so discard pkt */
		return DISCARD_PKT;
	}

	/*
	 * It's in the override table so check the destination port.  The
	 * override table index is offset by one so decrement i by one first.
	 */
	for (i--,j=1; j <= active.overrideTable[i].port[0].begin; j++) {
		if ((dstPort >= active.overrideTable[i].port[j].begin) &&
		    (dstPort <= active.overrideTable[i].port[j].end))
	    		return ALLOW_PKT;
	}

	/*
	 * If we made it this far, the destination port is not allowed
	 */
	DBstats.f_insideTcpPortOverride++;
	syslogMessage(SYSL_OUT_OVERRIDE, IPPROTO_TCP,
		      srcAddr, dstAddr, (int)srcPort, (int)dstPort);

	/* Check the TCP reset table */
	for (i=1; i <= active.t_rst_out[0].begin; i++) {
		if ((dstPort >= active.t_rst_out[i].begin) &&
		    (dstPort <= active.t_rst_out[i].end)) {
			DBstats.insideTCPreset++;
			return DISCARD_AND_ANSWER_PKT;
		}
	}
	return DISCARD_PKT;
}
	
/* ------------------------------------------------------------------------ */
/*
 * Parameters are expected to be in host byte order
 */

static int
checkIncomingUdp(u_int32_t srcAddr, u_int32_t dstAddr,
		 u_int16_t srcPort, u_int16_t dstPort)
{
	register int i;
	int class;
	PortListEntry *portList;

	/*
	 * Check for multicast
	 */
	if (IN_CLASSD(dstAddr)) {
		if (DBcfg.discardMulticast) {
			DBstats.f_outsideUdpMcast++;
			syslogMessage(SYSL_IN_CLASSD,IPPROTO_UDP,
				srcAddr, dstAddr, (int)srcPort, (int)dstPort);
			return DISCARD_PKT;
		}
		return ALLOW_PKT;
	}

	/*
	 * Do the lookup to get the class.
	 */
	class = treeLookup(active.addrTree, dstAddr, active.num_classes);

	/*
	 * See if the destination port is allowed by the 'udp' port list.
	 */
	portList = &active.u_dst_in[class * LIST_SIZE];
	for (i=1; i <= portList[0].begin; i++) {
		if ((dstPort >= portList[i].begin) &&
		    (dstPort <= portList[i].end))
			return ALLOW_PKT;
	}

        /*
         * If the destination port is not allowed then check the source port.
         */
	if (dstPort >= 1024) {
		portList = &active.u_src_in[class * LIST_SIZE];
		for (i=1; i <= portList[0].begin; i++) {
			if ((srcPort >= portList[i].begin) &&
			    (srcPort <= portList[i].end))
				return ALLOW_PKT;
		}
	}

	/*
	 * If we made it this far, the destination port is not allowed
	 */
	DBstats.f_outsideUdpPort++;
	syslogMessage(SYSL_IN_PORT, IPPROTO_UDP,
		      srcAddr, dstAddr, (int)srcPort, (int)dstPort);

	return DISCARD_PKT;
}

/* ------------------------------------------------------------------------ */
/*
 * Parameters are expected to be in host byte order
 */

static int
checkOutgoingUdp(u_int32_t srcAddr, u_int32_t dstAddr,
		 u_int16_t srcPort, u_int16_t dstPort)
{
	register int i;
	int class;
	PortListEntry *portList;

	/*
	 * Check for multicast
	 */
	if (IN_CLASSD(dstAddr)) {
		if (DBcfg.discardMulticast) {
			DBstats.f_insideUdpMcast++;
			syslogMessage(SYSL_OUT_CLASSD,IPPROTO_UDP,
				srcAddr, dstAddr, (int)srcPort, (int)dstPort);
			return DISCARD_PKT;
		}
		return ALLOW_PKT;
	}

	/*
	 * Do the lookup to get the class.
	 */
	class = treeLookup(active.addrTree, srcAddr, active.num_classes);

	/*
	 * See if the destination port is allowed by the 'udp' port list.
	 */
	portList = &active.u_dst_out[class * LIST_SIZE];
	for (i=1; i <= portList[0].begin; i++) {
		if ((dstPort >= portList[i].begin) &&
		    (dstPort <= portList[i].end))
			return ALLOW_PKT;
	}

        /*
         * If the destination port is not allowed then check the source port.
         */
	if (dstPort >= 1024) {
		portList = &active.u_src_out[class * LIST_SIZE];
		for (i=1; i <= portList[0].begin; i++) {
			if ((srcPort >= portList[i].begin) &&
			    (srcPort <= portList[i].end))
				return ALLOW_PKT;
		}
	}

	/*
	 * If we made it this far, the destination port is not allowed
	 */
	DBstats.f_insideUdpPort++;
	syslogMessage(SYSL_OUT_PORT, IPPROTO_UDP,
		      srcAddr, dstAddr, (int)srcPort, (int)dstPort);

	return DISCARD_PKT;
}

/* ------------------------------------------------------------------------ */
/*
 * Parameters are expected to be in host byte order
 */

static int
checkIncomingIcmp(u_int32_t srcAddr, u_int32_t dstAddr, u_int8_t type)
{
	register int i;
	PortListEntry *portList;

	portList = &active.i_typ_in[ treeLookup(active.addrTree,
				dstAddr, active.num_classes) * LIST_SIZE ];

	/*
	 * See if the ICMP type is allowed by the 'icmp' port list.
	 */
	for (i=1; i <= portList[0].begin; i++) {
		if ((type >= portList[i].begin) &&
		    (type <= portList[i].end))
			return ALLOW_PKT;
	}

	/*
	 * If we made it here, the ICMP type is not allowed
	 */
	DBstats.f_outsideIcmpType++;
	syslogMessage(SYSL_IN_TYPE, srcAddr, dstAddr, (int)type-1);

	return DISCARD_PKT;
}

/* ------------------------------------------------------------------------ */
/*
 * Parameters are expected to be in host byte order
 */

static int
checkOutgoingIcmp(u_int32_t srcAddr, u_int32_t dstAddr, u_int8_t type)
{
	register int i;
	PortListEntry *portList;

	portList = &active.i_typ_out[ treeLookup(active.addrTree,
				srcAddr, active.num_classes) * LIST_SIZE ];

	/*
	 * See if the ICMP type is allowed by the 'icmp' port list.
	 */
	for (i=1; i <= portList[0].begin; i++) {
		if ((type >= portList[i].begin) &&
		    (type <= portList[i].end))
			return ALLOW_PKT;
	}

	/*
	 * If we made it here, the ICMP type is not allowed
	 */
	DBstats.f_insideIcmpType++;
	syslogMessage(SYSL_OUT_TYPE, srcAddr, dstAddr, (int)type-1);

	return DISCARD_PKT;
}

/* ------------------------------------------------------------------------ */

static int
checkRejectTable(struct ip *ipHeader, u_int32_t srcAddr, u_int32_t dstAddr)
{
	/* If the reject table is empty, then forward all packets */
	if (active.rejectTree == NULL)
		return ALLOW_PKT;

	/* If srcAddr is in the tree and it's class is 1, then discard pkt */
	if (treeLookup(active.rejectTree, srcAddr, 2) == 1) {
		DBstats.f_outsideRejectTable++;
		syslogMessage(SYSL_IN_REJECT, (int)ipHeader->ip_p,
			srcAddr, dstAddr);
		return DISCARD_PKT;
	}

	/* Process by default */
	return ALLOW_PKT;
}

/* ------------------------------------------------------------------------ */
/*
 * Parameters 'protocol' and 'protolen' are expected to be in host byte order
 * The packet pointed to by 'pkt' is, of course, in network btye order
 */

static int
checkIncomingPacket(u_int16_t protocol, u_int8_t *pkt,
		    u_int protolen, int checkAccept)
{
	struct ip *ipHeader;
	struct tcphdr *tcpHeader;
	struct udphdr *udpHeader;
	struct icmp *icmpHeader;
	u_int32_t srcAddr, dstAddr;
	u_int16_t ip_len;

	switch (protocol) {

	case ETHERTYPE_IP:
		ipHeader = (struct ip *)pkt;
		srcAddr = ntohl(ipHeader->ip_src.s_addr);
		dstAddr = ntohl(ipHeader->ip_dst.s_addr);

		if (checkAccept) {
			if (checkAcceptTable(ipHeader, srcAddr, dstAddr)
			    == DISCARD_PKT)
				return DISCARD_PKT;
		} else {
			if (checkRejectTable(ipHeader, srcAddr, dstAddr)
			    == DISCARD_PKT)
				return DISCARD_PKT;
		}

		/*
		 * Pass all IP fragments that do not have offset 0 (beginning
		 * of the packet) without checking since the TCP/UDP
		 * headers are not in this packet.
		 */
		if ((ipHeader->ip_off & SW_IP_OFFMASK) != 0) {
			/*
			 * If option is set, then discard pkts with offset of 1
			 */
			if (DBcfg.discardSuspectOffset &&
			    (ntohs(ipHeader->ip_off & SW_IP_OFFMASK) == 1) &&
			    (ipHeader->ip_p == IPPROTO_TCP)) {

				DBstats.f_outsideSuspectOffset++;
				syslogMessage(SYSL_IN_OFFSET, IPPROTO_TCP,
					srcAddr, dstAddr);
				return DISCARD_PKT;
			}

			/*
			 * If option is set, then discard fragmented ICMP pkts
			 */
			if (DBcfg.discardFragmentedICMP &&
			    (ipHeader->ip_p == IPPROTO_ICMP)) {

				DBstats.f_outsideIcmpFrag++;
			    	syslogMessage(SYSL_IN_FRAG, IPPROTO_ICMP,
					srcAddr, dstAddr);
				return DISCARD_PKT;
			}

			return ALLOW_PKT;
		}

		switch (ipHeader->ip_p) {

		case IPPROTO_TCP:
			tcpHeader = (struct tcphdr *)(pkt + 
						(ipHeader->ip_hl << 2)); 
			/*
			 * Make sure this packet (fragment) includes enough of
			 * the TCP header before making the other checks.
			 * Otherwise a SYN can sneak through since we might be
			 * trying to test something which is actually in the
			 * next fragment.
			 *
			 * NOTE: we drop all of these since we do not keep any
			 *       state between fragments.
			 *
			 * Many thanks to Uwe Ellermann at DFN-CERT for
			 * reporting this problem.
			 */
			ip_len = ntohs(ipHeader->ip_len);
			if ((ip_len > protolen) ||
			    ((ip_len - (ipHeader->ip_hl << 2)) < 14)) {
				DBstats.f_outsideShortTcpHdr++;
				syslogMessage(SYSL_IN_LENGTH, IPPROTO_TCP,
					srcAddr, dstAddr,
					ntohs(tcpHeader->th_sport),
					ntohs(tcpHeader->th_dport));

				return DISCARD_PKT;
			}

			/* Check for "ACKless SYN" */
			if ((tcpHeader->th_flags & (TH_SYN|TH_ACK)) == TH_SYN) {
				return checkIncomingTcp(srcAddr, dstAddr,
						ntohs(tcpHeader->th_sport),
						ntohs(tcpHeader->th_dport));
			}
			break;

		case IPPROTO_UDP:
			udpHeader = (struct udphdr *)(pkt + 
						(ipHeader->ip_hl << 2));
			/*
			 * Make sure this packet (fragment) includes enough
			 * of the UDP header before making the other checks.
			 */
			ip_len = ntohs(ipHeader->ip_len);
			if ((ip_len > protolen) ||
			    ((ip_len - (ipHeader->ip_hl << 2)) < 4)) {
				DBstats.f_outsideShortUdpHdr++;
				syslogMessage(SYSL_IN_LENGTH, IPPROTO_UDP,
					srcAddr, dstAddr,
					ntohs(udpHeader->uh_sport),
					ntohs(udpHeader->uh_dport));

				return DISCARD_PKT;
			}

			/* Check UDP protocols. */

			return checkIncomingUdp(srcAddr, dstAddr,
						ntohs(udpHeader->uh_sport),
						ntohs(udpHeader->uh_dport));
			break;

		case IPPROTO_ICMP:
			icmpHeader = (struct icmp *)
				     (pkt + (ipHeader->ip_hl << 2));

			/*
			 * If option is set, then discard fragmented ICMP pkts
			 */
			if (DBcfg.discardFragmentedICMP &&
			    (ipHeader->ip_off & IP_MF) != 0) {
				DBstats.f_outsideIcmpFrag++;
			    	syslogMessage(SYSL_IN_FRAG, IPPROTO_ICMP,
					srcAddr, dstAddr);

				return DISCARD_PKT;
			}

			/*
			 * If option is set, discard attack ICMP pkts.
			 *
			 * If the ICMP type, code, identifier, and seq num all
			 * equal 0, this may be an ICMP echo reply attack pkt.
			 */
			if (DBcfg.discardAttackICMP &&
			    (*((u_int16_t *)icmpHeader) == 0) && /*type & code*/
			    (icmpHeader->icmp_void == 0)) {      /*id & seq*/

				if (icmpHeader->icmp_cksum == 0xFFFF) {
				
					/* This is a "smurf" attack */
					DBstats.f_outsideSmurfDos++;
			    		syslogMessage(SYSL_IN_DOS,
						      srcAddr, dstAddr,
						      DOS_SMURF);
					return DISCARD_PKT;

				} else if (*((u_int16_t *)icmpHeader->icmp_data)
					   == htons(0x4500)) {
					/* This is a "pong" attack */
					DBstats.f_outsidePongDos++;
			    		syslogMessage(SYSL_IN_DOS,
						      srcAddr, dstAddr,
						      DOS_PONG);
					return DISCARD_PKT;
				}
			}

			return checkIncomingIcmp(srcAddr, dstAddr,
						 icmpHeader->icmp_type);
			break;

		default:
			if (DBcfg.discardOtherIp) {

				DBstats.f_outsideOtherIp++;
				syslogMessage(SYSL_IN_PROT, (int)ipHeader->ip_p,
					srcAddr, dstAddr);

				return DISCARD_PKT;
			}

			/* Everything is allowed so far. */
			break;
		}

		break;

	case ETHERTYPE_ARP:
	case ETHERTYPE_REVARP:
		/* pass these guys through no matter what */
		break;

	default:
		/* Only pass the rest if told to do so. */
		if (DBcfg.discardNonIp) {

			/* Boy will this get ugly if the syslog mask is not */
			/* configured properly..... */

			DBstats.f_outsideNonIp++;
			syslogMessage(SYSL_IN_FILTER, protocol);
			return DISCARD_PKT;
		}
	}

	return ALLOW_PKT;
}

/* ------------------------------------------------------------------------ */

static int
checkAcceptTable(struct ip *ipHeader, u_int32_t srcAddr, u_int32_t dstAddr)
{
	/* If the accept table is empty, then forward all packets */
	if (active.acceptTree == NULL)
		return ALLOW_PKT;

	/* If srcAddr is in the tree and it's class is 1, then process pkt */
	if (treeLookup(active.acceptTree, srcAddr, 2) == 1)
		return ALLOW_PKT;

	/* Discard by default */
	DBstats.f_insideAcceptTable++;
	syslogMessage(SYSL_OUT_ACCEPT, (int)ipHeader->ip_p,
			srcAddr, dstAddr);
	return DISCARD_PKT;
}

/* ------------------------------------------------------------------------ */
/*
 * Parameters 'protocol' and 'protolen' are expected to be in host byte order
 * The packet pointed to by 'pkt' is, of course, in network btye order
 */

static int
checkOutgoingPacket(u_int16_t protocol, u_int8_t *pkt, u_int protolen)
{
	struct ip *ipHeader;
	struct tcphdr *tcpHeader;
	struct udphdr *udpHeader;
	struct icmp *icmpHeader;
	u_int16_t ip_len;
	u_int32_t srcAddr, dstAddr;

	switch (protocol) {

	case ETHERTYPE_IP:
		ipHeader = (struct ip *)pkt;
		srcAddr = ntohl(ipHeader->ip_src.s_addr);
		dstAddr = ntohl(ipHeader->ip_dst.s_addr);

		if (checkAcceptTable(ipHeader,srcAddr,dstAddr) == DISCARD_PKT)
			return DISCARD_PKT;

		/*
		 * Pass all IP fragments that do not have offset 0 (beginning
		 * of the packet) without checking since the TCP/UDP
		 * headers are not in this packet.  An area for improvement
		 * would be to cache session info so we could drop all
		 * disallowed fragments also instead of just the first one.
		 */
		if ((ipHeader->ip_off & SW_IP_OFFMASK) != 0) {

			if (DBcfg.discardSuspectOffset &&
			    (ntohs(ipHeader->ip_off & SW_IP_OFFMASK) == 1) &&
			    (ipHeader->ip_p == IPPROTO_TCP)) {

				DBstats.f_insideSuspectOffset++;
				syslogMessage(SYSL_OUT_OFFSET, IPPROTO_TCP,
					srcAddr, dstAddr);

				return DISCARD_PKT;
			}

			/*
			 * If option is set, then discard fragmented ICMP pkts
			 */
			if (DBcfg.discardFragmentedICMP &&
			    (ipHeader->ip_p == IPPROTO_ICMP)) {

				DBstats.f_insideIcmpFrag++;
			    	syslogMessage(SYSL_OUT_FRAG, IPPROTO_ICMP,
					srcAddr, dstAddr);
				return DISCARD_PKT;
			}

			return ALLOW_PKT;
		}

		switch (ipHeader->ip_p) {

		case IPPROTO_TCP:
			tcpHeader = (struct tcphdr *)(pkt + 
						(ipHeader->ip_hl << 2)); 
			/*
			 * Make sure this packet (fragment) includes enough
			 * of the TCP header before making the other checks.
			 */
			ip_len = ntohs(ipHeader->ip_len);
			if ((ip_len > protolen) ||
			    ((ip_len - (ipHeader->ip_hl << 2)) < 14)) {
				DBstats.f_insideShortTcpHdr++;
				syslogMessage(SYSL_OUT_LENGTH, IPPROTO_TCP,
					srcAddr, dstAddr,
					ntohs(tcpHeader->th_sport),
					ntohs(tcpHeader->th_dport));

				return DISCARD_PKT;
			}

			/* Check for "ACKless SYN" */
			if ((tcpHeader->th_flags & (TH_SYN|TH_ACK)) == TH_SYN) {
				return checkOutgoingTcp(srcAddr, dstAddr,
						ntohs(tcpHeader->th_sport),
						ntohs(tcpHeader->th_dport));
			}
			break;

		case IPPROTO_UDP:
			udpHeader = (struct udphdr *)(pkt + 
						(ipHeader->ip_hl << 2));
			/*
			 * Make sure this packet (fragment) includes enough
			 * of the UDP header before making the other checks.
			 */
			ip_len = ntohs(ipHeader->ip_len);
			if ((ip_len > protolen) ||
			    ((ip_len - (ipHeader->ip_hl << 2)) < 4)) {
				DBstats.f_insideShortUdpHdr++;
				syslogMessage(SYSL_OUT_LENGTH, IPPROTO_UDP,
					srcAddr, dstAddr,
					ntohs(udpHeader->uh_sport),
					ntohs(udpHeader->uh_dport));

				return DISCARD_PKT;
			}

			return  checkOutgoingUdp( srcAddr, dstAddr,
					ntohs(udpHeader->uh_sport),
					ntohs(udpHeader->uh_dport));
			break;

		case IPPROTO_ICMP:
			icmpHeader = (struct icmp *)
				     (pkt + (ipHeader->ip_hl << 2));

			/*
			 * If option is set, then discard fragmented ICMP pkts
			 */
			if (DBcfg.discardFragmentedICMP &&
			    (ipHeader->ip_off & IP_MF) != 0) {
				DBstats.f_insideIcmpFrag++;
			    	syslogMessage(SYSL_OUT_FRAG, IPPROTO_ICMP,
					srcAddr, dstAddr);

				return DISCARD_PKT;
			}

			/*
			 * If option is set, discard attack ICMP pkts.
			 *
			 * If the ICMP type, code, identifier, and seq num all
			 * equal 0, this may be an ICMP echo reply attack pkt.
			 */
			if (DBcfg.discardAttackICMP &&
			    (*((u_int16_t *)icmpHeader) == 0) && /*type & code*/
			    (icmpHeader->icmp_void == 0)) {      /*id & seq*/

				if (icmpHeader->icmp_cksum == 0xFFFF) {
				
					/* This is a "smurf" attack */
					DBstats.f_insideSmurfDos++;
			    		syslogMessage(SYSL_OUT_DOS,
						      srcAddr, dstAddr,
						      DOS_SMURF);
					return DISCARD_PKT;

				} else if (*((u_int16_t *)icmpHeader->icmp_data)
					   == htons(0x4500)) {
					/* This is a "pong" attack */
					DBstats.f_insidePongDos++;
			    		syslogMessage(SYSL_OUT_DOS,
						      srcAddr, dstAddr,
						      DOS_PONG);
					return DISCARD_PKT;
				}
			}

			return checkOutgoingIcmp(srcAddr, dstAddr,
						 icmpHeader->icmp_type);
			break;

		default:
			if (DBcfg.discardOtherIp) {

				DBstats.f_insideOtherIp++;
				syslogMessage(SYSL_OUT_PROT,(int)ipHeader->ip_p,
					srcAddr, dstAddr);

				return DISCARD_PKT;
			}

			/* Everything is allowed so far. */
			break;
		}

		break;

	case ETHERTYPE_ARP:
	case ETHERTYPE_REVARP:
		/* pass these guys through no matter what */
		break;

	default:
		/* Only pass the rest if told to do so. */
		if (DBcfg.discardNonIp) {

			DBstats.f_insideNonIp++;
			syslogMessage(SYSL_OUT_FILTER, (int)protocol);
			return DISCARD_PKT;
		}
	}

	return ALLOW_PKT;
}

/* ------------------------------------------------------------------------ */
/*
 * This is only called for packets that are originated by drawbridge.
 * Parameter 'dst' is expected to be in network byte order
 */

int
drawbridge_bridge_lookup(HardwareAddr *dst, struct ifnet *ifp)
{
	register int hash;
	register int curr;

	if (!DBstats.running ||
	    ((ifp != DBcfg.in_ifp) && (ifp != DBcfg.out_ifp)))
	    	/*
		 * Drawbridge is not running or this is not one of
		 * our interfaces so just process the pkt normally.
		 */
		return DB_PROCESS_PKT;

	/*
	 * If broadcast, then the packet needs to go to any interface
	 * that is setup for 'listen'.  If neither is enabled for
	 * 'listen', then drop the packet.
	 */
	if (IS_BROADCAST(dst))
		return DBcfg.listenMode;

	/*
	 * Check the bridge table for the destination address.
	 */
	hash = (dst->words[0] ^ dst->words[1] ^ dst->words[2])
		& ADDRESS_HASH_MASK;

	for (curr = hash; hashTable[curr].valid; ) {

		if (IS_HW_ADDR_EQU(dst, hashTable[curr].address)) {
		    	/*
			 * We found the address and know the side it's on.
			 * If that interface is enabled for 'listen', then
			 * send the packet.  Otherwise, drop it.
			 */
			return (hashTable[curr].interface & DBcfg.listenMode);
		}

		/*
		 * Increment 'curr' and check to see if we have wrapped.
		 * If so, the table is full and the address was not found.
		 */
		if ((curr = (curr+1) & ADDRESS_HASH_MASK) == hash)
			break;
	}

	/*
	 * Not found.  We don't know which side the destination is on so
	 * send the packet out the interface(s) that have 'listen' enabled.
	 */
	return DBcfg.listenMode;
} 

/* ------------------------------------------------------------------------ */
/*
 * Main bridge function - called for packets not originated by drawbridge.
 * Parameters 'src' and 'dst' are expected to be in network byte order
 */

static int
bridge(int interface, HardwareAddr *src, HardwareAddr *dst)
{
	int pass = YES;

	/*
	 * Check if the destination address is a broadcast or
	 * multicast address and always forward if so.
	 */
	if (!IS_BROADCAST(dst)) {
		register int curr;
		register int hash;

		/*
		 * Check the bridge table for the destination address.
		 */
		hash = (dst->words[0] ^ dst->words[1] ^ dst->words[2])
			& ADDRESS_HASH_MASK;

		for (curr = hash; hashTable[curr].valid; ) {

			if (IS_HW_ADDR_EQU(dst, hashTable[curr].address)) {
			    	/*
				 * We found the address and know which side of
				 * the bridge it lives on.  Only forward if
				 * this packet is for a host on the other side.
				 */
				if (interface == hashTable[curr].interface)
					pass = NO;
				break;
			}

			/*
			 * Increment 'curr' and check to see if we have wrapped.
			 * If so, the table is full and the address was not
			 * found so forward the packet regardless.
			 */
			if ((curr = (curr+1) & ADDRESS_HASH_MASK) == hash)
				break;
		}
	}

	/*
	 * Now check the source address and make sure it's in the bridge
	 * table.  Checking for broadcast is somewhat superfluous but if
	 * some broken implementation sources a broadcast or multicast
	 * address, we sure don't want to add it to the bridge table!
	 */
	if (!IS_BROADCAST(src)) {
		register int curr;
		register int hash;

		/*
		 * Check the bridge table for the source address.
		 */
		hash = (src->words[0] ^ src->words[1] ^ src->words[2])
			& ADDRESS_HASH_MASK;

		for (curr = hash; ; ) {

			if (IS_HW_ADDR_EQU(src, hashTable[curr].address)) {
				u_int s;
				
			    	/*
				 * It's already in the table so just make
				 * sure that the side of the bridge it's on
				 * stays up to date.
				 */
				s = splimp();
				
				hashTable[curr].interface = interface;
				
				splx(s);
			    	break;
			}

			/*
			 * If the above compare was false, and this slot is
			 * open, then the address is not in the table and
			 * we can just instert it here.
			 */
			if (hashTable[curr].valid == NO) {
				u_int s;

				s = splimp();
				
                                hashTable[curr].address.addr = src->addr;
                                hashTable[curr].interface = interface;
                                hashTable[curr].valid = YES;

				splx(s);

				DBstats.bridgeEntries++;
				break;
			}

			/*
			 * Increment 'curr' and check to see if we have wrapped.
			 * If so, then the table is already full.
			 */
			if ((curr = (curr+1) & ADDRESS_HASH_MASK) == hash)
				break;
		}
	}

	return pass;
}

/* ------------------------------------------------------------------------ */
/*
 * 'ifp' is a pointer to the interface pointer
 * 'h' is a pointer to the packet header
 * 'm' is a pointer to the mbuf containing the packet
 */

int
drawbridge_test_pkt(struct ifnet **ifp, void *h, struct mbuf *m)
{
	u_int16_t protocol;
	u_int8_t *pkt;
	u_int protolen, pktlen;
	HardwareAddr *dhost;
	HardwareAddr *shost;

	if (!DBstats.running)
		return DB_PROCESS_PKT;

	switch (DBcfg.media_type) {
	case IFT_ETHER:
		dhost = (HardwareAddr *)((struct ether_header *)h)->ether_dhost;
		shost = (HardwareAddr *)((struct ether_header *)h)->ether_shost;
		protocol = ntohs(((struct ether_header *)h)->ether_type);
		pkt = m->m_data;
		protolen = m->m_pkthdr.len;
		pktlen = m->m_pkthdr.len + sizeof(struct ether_header);
		break;

	case IFT_FDDI:
		dhost = (HardwareAddr *)((struct fddi_header *)h)->fddi_dhost;
		shost = (HardwareAddr *)((struct fddi_header *)h)->fddi_shost;
		if (*((u_char *)(m->m_data)) == LLC_SNAP_LSAP) {
			protocol = ntohs(mtod(m, struct llc *)->
					 llc_snap.ether_type);
                	pkt = m->m_data + LLC_SNAPLEN;
			protolen = m->m_pkthdr.len - LLC_SNAPLEN;
		} else {
			protocol = 0;
                	pkt = m->m_data;
			protolen = m->m_pkthdr.len;
		}
		pktlen = m->m_pkthdr.len + sizeof(struct fddi_header);
		break;

	default:
		printf("unsupported media - disabling drawbridge\n");
		DBstats.running = NO;
		return DB_PROCESS_PKT;
	}

	if (*ifp == DBcfg.in_ifp) {   /* pkt from the inside interface */
		register int rc;

		DBstats.p_insideRx++;
		DBstats.b_insideRx += pktlen;

		/*
		 * Check the bridge table to see if the destination host is
		 * on the same side as the source host.  If so, drop the pkt.
		 */
		if (bridge(DB_INSIDE_IF, shost, dhost) == NO)
			return DB_DROP_PKT | mirror_i1;

		/*
		 * The packet should be delivered locally if it is directed
		 * to us or if it is an ARP for us.  If it is for us and
		 * listenMode is disabled for the interface, drop the pakcet.
		 */
		if (IS_HW_ADDR_EQU(dhost, DBcfg.local_hwaddr) ||
		    (IS_BROADCAST(dhost) && (protocol == ETHERTYPE_ARP) &&
		    (DBcfg.local_IP.s_addr == ARP_TARGET_IP_ADDR(pkt)))) {
			if (DBcfg.listenMode & DB_INSIDE_IF) {
				/* Check the incoming packet filters */
				if (checkIncomingPacket(protocol, pkt,
				    protolen, 1) == DISCARD_PKT) {
					DBstats.p_outsideFiltered++;
					DBstats.b_outsideFiltered += pktlen;
					return DB_DROP_PKT | mirror_i2;
				}
				*ifp = DBcfg.local_ifp;
				m->m_pkthdr.rcvif = DBcfg.local_ifp;
				return DB_PROCESS_PKT | mirror_i1;
			} else
				return DB_DROP_PKT | mirror_i2;
		}

		/*
		 * Check to see if the packet should be allowed through.
		 */
		rc = checkOutgoingPacket(protocol,pkt,protolen);
		if (rc != ALLOW_PKT) {
			DBstats.p_insideFiltered++;
			DBstats.b_insideFiltered += pktlen;
			if (rc == DISCARD_AND_ANSWER_PKT)
				return DB_DROP_AND_ANSWER_PKT | mirror_i2;
			return DB_DROP_PKT | mirror_i2;
		}

		/*
		 * Forward the pkt to the outside interface
		 */
		DBstats.p_outsideTx++;
		DBstats.b_outsideTx += pktlen;
		*ifp = DBcfg.out_ifp;
		return DB_FORWARD_PKT | mirror_b;

	} else if (*ifp == DBcfg.out_ifp) {  /* from the outside interface */
		register int rc;

		DBstats.p_outsideRx++;
		DBstats.b_outsideRx += pktlen;

		/*
		 * Check the bridge table to see if the destination host is
		 * on the same side as the source host.  If so, drop the pkt.
		 */
		if (bridge(DB_OUTSIDE_IF, shost, dhost) == NO)
			return DB_DROP_PKT | mirror_o1;

		/*
		 * The packet should be delivered locally if it is directed
		 * to us or if it is an ARP for us.  If it is for us and
		 * listenMode is disabled for the interface, drop the pakcet.
		 */
		if (IS_HW_ADDR_EQU(dhost, DBcfg.local_hwaddr) ||
		    (IS_BROADCAST(dhost) && (protocol == ETHERTYPE_ARP) &&
		    (DBcfg.local_IP.s_addr == ARP_TARGET_IP_ADDR(pkt)))) {
			if (DBcfg.listenMode & DB_OUTSIDE_IF) {
				*ifp = DBcfg.local_ifp;
				m->m_pkthdr.rcvif = DBcfg.local_ifp;

				/* Check the incoming packet filters */
				if (checkIncomingPacket(protocol, pkt,
				    protolen, 0) == DISCARD_PKT) {
					DBstats.p_outsideFiltered++;
					DBstats.b_outsideFiltered += pktlen;
					return DB_DROP_PKT | mirror_o2;
				}
				return DB_PROCESS_PKT | mirror_o1;
			} else
				return DB_DROP_PKT | mirror_o2;
		}

		/*
		 * Check to see if the packet should be allowed through.
		 */
		rc = checkIncomingPacket(protocol,pkt,protolen,0);
		if (rc != ALLOW_PKT) {
			DBstats.p_outsideFiltered++;
			DBstats.b_outsideFiltered += pktlen;
			if (rc == DISCARD_AND_ANSWER_PKT)
				return DB_DROP_AND_ANSWER_PKT | mirror_o2;
			return DB_DROP_PKT | mirror_o2;
		}

		/*
		 * forward the pkt to the inside interface
		 */
		DBstats.p_insideTx++;
		DBstats.b_insideTx += pktlen;
		*ifp = DBcfg.in_ifp;
		return DB_FORWARD_PKT | mirror_b;
	}

	/*
	 * The pkt is not from either filter interface so process it normally
	 */
	return DB_PROCESS_PKT;
}

/* ------------------------------------------------------------------------ */
/*
 * Sends tcp reset packet.
 * 'r' is a pointer to the mbuf containing the outbound header and packet
 * 'm' is a pointer to the mbuf containing the inbound packet
 * 'h' is a pointer to the inbound packet media header
 */

void
drawbridge_response_pkt(struct mbuf *r, void *h, struct mbuf *m)
{
	u_int32_t tmp;
	struct pkt_struct {
		struct ip i;
		struct tcphdr t;
	};
	register struct pkt_struct *out_pkt;
	register struct pkt_struct *in_pkt;

	/* Outbound packet's media header */
	if (DBcfg.media_type == IFT_ETHER) {
		register struct ether_header *in_h = h;
		register struct ether_header *out_h =
			mtod(r, struct ether_header *);

		memcpy(out_h->ether_dhost, in_h->ether_shost, 6);
		memcpy(out_h->ether_shost, in_h->ether_dhost, 6);
		out_h->ether_type = in_h->ether_type;

		r->m_data += sizeof(struct ether_header);

	} else { /* IFT_FDDI */
		register struct fddi_header *in_h = h;
		register struct fddi_header *out_h =
			mtod(r, struct fddi_header *);

		memcpy(out_h->fddi_dhost, in_h->fddi_shost, 6);
		memcpy(out_h->fddi_shost, in_h->fddi_dhost, 6);
		out_h->fddi_fc = in_h->fddi_fc;

		r->m_data += sizeof(struct fddi_header);
		memcpy(r->m_data, m->m_data, LLC_SNAPLEN);
		r->m_data += LLC_SNAPLEN;
		m->m_data += LLC_SNAPLEN;
	}

	/* Setup pointers and clear outbound packet */
	in_pkt = mtod(m, struct pkt_struct *);
	out_pkt = mtod(r, struct pkt_struct *);
	r->m_len = sizeof(struct pkt_struct);
	bzero(out_pkt, r->m_len);

	/* IP header fields included in the TCP checksum */
	out_pkt->i.ip_len = htons(20);		/* TCP len */
	out_pkt->i.ip_p = IPPROTO_TCP;
	out_pkt->i.ip_src = in_pkt->i.ip_dst;
	out_pkt->i.ip_dst = in_pkt->i.ip_src;

	/* TCP header */
	out_pkt->t.th_sport = in_pkt->t.th_dport;
	out_pkt->t.th_dport = in_pkt->t.th_sport;
	out_pkt->t.th_seq = in_pkt->t.th_ack;
	tmp = ntohl(in_pkt->t.th_seq) + 1;
	out_pkt->t.th_ack = htonl(tmp);
	out_pkt->t.th_off = 5;
	out_pkt->t.th_flags = 0x14;

	/* TCP checksum */
	out_pkt->t.th_sum = in_cksum(r, r->m_len);

	/* Finish the IP header */
	out_pkt->i.ip_v = 4;
	out_pkt->i.ip_hl = 5;			/* IP hdr len >> 2 */
	out_pkt->i.ip_len = htons(40);		/* Total length */
	out_pkt->i.ip_ttl = 128;

	/* IP header checksum */
	out_pkt->i.ip_sum = in_cksum_hdr(&out_pkt->i);

	/* Adjust outbound mbuf to include media header */
	if (DBcfg.media_type == IFT_ETHER) {
		r->m_data -= sizeof(struct ether_header);
		r->m_len += sizeof(struct ether_header);
	} else {
		r->m_data -= (sizeof(struct fddi_header) + LLC_SNAPLEN);
		r->m_len += (sizeof(struct fddi_header) + LLC_SNAPLEN);
	}
	r->m_pkthdr.len = r->m_len;
}

/* ------------------------------------------------------------------------ */

static int
db_open(dev_t dev, int flags, int fmt, struct proc *p)
{
	if (minor(dev) >= 1)
		return ENXIO;
#if defined(__FreeBSD__) && __FreeBSD__ >= 4
        make_dev(&db_cdevsw, 0, UID_ROOT, GID_WHEEL, 0600, "drawbridge");
#endif
	return NO_ERROR;
}

/* ------------------------------------------------------------------------ */

static int
db_close(dev_t dev, int flags, int fmt, struct proc *p)
{
	if (minor(dev) >= 1)
		return ENXIO;

	return NO_ERROR;
}

/* ------------------------------------------------------------------------ */

static int
db_ifpromisc(struct ifnet *ifp, int pswitch)
{
        struct ifreq ifr;
	u_int s;
	int r;

	s = splimp();
        if (pswitch) {
                if ((ifp->if_flags & IFF_UP) == 0)
                        return ERROR_IF_DOWN;
                if (ifp->if_pcount++ != 0)
                        return NO_ERROR;
                ifp->if_flags |= IFF_PROMISC;
        } else {
                if (--ifp->if_pcount > 0)
                        return NO_ERROR;
                ifp->if_flags &= ~IFF_PROMISC;
        }
        ifr.ifr_flags = ifp->if_flags;
        r = ((*ifp->if_ioctl)(ifp, SIOCSIFFLAGS, (caddr_t)&ifr));
	splx(s);

	return r;
}

/* ------------------------------------------------------------------------ */

static int
db_gethwaddr(struct ifnet *ifp, HardwareAddr *addr)
{
	/* Assume that ifp is actually pointer to arpcom structure */
	register struct arpcom *ac = (struct arpcom *)ifp;

	if (ac == NULL)
		return ERROR_GET_HW;

	bcopy(ac->ac_enaddr, (caddr_t)addr, sizeof(*addr));

	return NO_ERROR;
}

/* ------------------------------------------------------------------------ */

static int
db_getIPaddr(struct ifnet *ifp, in_addr_t *addr)
{
	struct in_ifaddr *ia;

	IFP_TO_IA(ifp, ia);

	if (ia == NULL)
		return ERROR_GET_IP;

	bcopy(&(ia->ia_addr.sin_addr), (caddr_t)addr, sizeof(*addr));

	return NO_ERROR;
}

/* ------------------------------------------------------------------------ */

static int
initRequest(initReq *req)
{
	int error;

	/*
	 * Clear all config info and stats (except log info)
	 */
	initialized = NO;
	bzero(&DBcfg.CFG_CLR_START, CFG_CLR_SIZE);
	bzero(&DBstats, sizeof(DBstats));
	mirror_i1 = mirror_i2 = mirror_o1 = mirror_o2 = mirror_b = 0;

	/*
	 * Map the interface names to interface pointers.  Also
	 * make sure the interfaces are valid and that the user
	 * didn't specify the same interface for both sides. 
	 */
	if ((DBcfg.in_ifp = ifunit(req->inName)) == NULL) {
		DB_PRINTF(("DIOCINIT: invalid interface (inside): %s\n",
			req->inName));
		return ERROR_INVALID_IF; /* invalid interface */
	}
	if ((DBcfg.out_ifp = ifunit(req->outName)) == NULL) {
		DB_PRINTF(("DIOCINIT: invalid interface (outside): %s\n",
			req->outName));
		return ERROR_INVALID_IF; /* invalid interface */
	}
	if (DBcfg.in_ifp == DBcfg.out_ifp) {
		DB_PRINTF(("DIOCINIT: can't use same interface: %s\n",
			req->inName));
		return ERROR_SAME_IF;
	}
	if (req->mirrorName[0]) {
		DBcfg.mirror_ifp = ifunit(req->mirrorName);
		if ((DBcfg.mirror_ifp == NULL) ||
		    (DBcfg.mirror_ifp == DBcfg.in_ifp) ||
		    (DBcfg.mirror_ifp == DBcfg.out_ifp)) {
			DB_PRINTF(("DIOCINIT: invalid interface (mirror): %s\n",
				req->mirrorName));
			return ERROR_INVALID_IF; /* invalid interface */
		}
	}

	/*
	 * check the media type of the interfaces
	 */
	DBcfg.media_type = DBstats.media_type = DBcfg.in_ifp->if_type;
	if ((DBcfg.out_ifp->if_type != DBcfg.media_type) || (DBcfg.mirror_ifp
	     && (DBcfg.mirror_ifp->if_type != DBcfg.media_type))) {
		DB_PRINTF(("DIOCINIT: can't use different media types\n"));
		return ERROR_DIFF_IF_TYPE;  /* must be same media*/
	}
	if ((DBcfg.media_type != IFT_ETHER) && (DBcfg.media_type != IFT_FDDI)) {
		DB_PRINTF(("DIOCINIT: unsupported media type\n"));
		return ERROR_BAD_IF_TYPE;  /* media not supported */
	}

	/*
	 * make sure interfaces are up and running
	 */
	if ((DBcfg.in_ifp->if_flags & (IFF_UP|IFF_RUNNING)) !=
	    (IFF_UP|IFF_RUNNING)) {
	    	DB_PRINTF(("DIOCINIT: inside interface down or not running\n"));
		return ERROR_INSIDE_DOWN;
	}
	if ((DBcfg.out_ifp->if_flags & (IFF_UP|IFF_RUNNING)) !=
	    (IFF_UP|IFF_RUNNING)) {
	    	DB_PRINTF(("DIOCINIT:outside interface down or not running\n"));
		return ERROR_OUTSIDE_DOWN;
	}
	if (DBcfg.mirror_ifp &&
	    ((DBcfg.mirror_ifp->if_flags & (IFF_UP|IFF_RUNNING)) !=
	     (IFF_UP|IFF_RUNNING))) {
	    	DB_PRINTF(("DIOCINIT:mirror interface down or not running\n"));
		return ERROR_MIRROR_DOWN;
	}

	/*
	 * get hw addr of interfaces
	 */
	if ((error = db_gethwaddr(DBcfg.in_ifp, &DBcfg.in_hwaddr))) {
		DB_PRINTF(("DIOCINIT: could not get hw addr (inside)\n"));
		return error;
	}
	if ((error = db_gethwaddr(DBcfg.out_ifp, &DBcfg.out_hwaddr))) {
		DB_PRINTF(("DIOCINIT: could not get hw addr (outside)\n"));
		return error;
	}

	/*
	 * get IP addr of interfaces
	 */
	if ((error = db_getIPaddr(DBcfg.in_ifp, &DBcfg.in_IP))) {
		DB_PRINTF(("DIOCINIT: could not get IP addr (inside)\n"));
		return error;
	}
	if ((error = db_getIPaddr(DBcfg.out_ifp, &DBcfg.out_IP))) {
		DB_PRINTF(("DIOCINIT: could not get IP addr (outside)\n"));
		return error;
	}

	/*
	 * Init other stuff
	 */
	if (ptrs_initialized == 0) {
		/* only do this the first time initRequest is called */
		bzero(&active, sizeof(pointerStruct));
		bzero(&new, sizeof(pointerStruct));
		ptrs_initialized = 1;
	}

	/* setup default filters */
	initTables(&active, 1);

	/* clear the bridge table */
	bzero(hashTable, sizeof(hashTable));

	strncpy(DBcfg.in_name,req->inName, MAX_DEV_NAMELEN);
	strncpy(DBcfg.out_name,req->outName, MAX_DEV_NAMELEN);
	strncpy(DBcfg.mirror_name,req->mirrorName, MAX_DEV_NAMELEN);

	DB_PRINTF(("DIOCINIT: inside if '%s' hw addr " \
		   "%02x:%02x:%02x:%02x:%02x:%02x\n", req->inName,
		   DBcfg.in_hwaddr.bytes[0], DBcfg.in_hwaddr.bytes[1],
		   DBcfg.in_hwaddr.bytes[2], DBcfg.in_hwaddr.bytes[3],
		   DBcfg.in_hwaddr.bytes[4], DBcfg.in_hwaddr.bytes[5]));
	DB_PRINTF(("DIOCINIT: inside if '%s' ip addr " \
		   "%d.%d.%d.%d\n", req->inName,
		   DBcfg.in_IP.s_un_b.s_b[0],
		   DBcfg.in_IP.s_un_b.s_b[1],
		   DBcfg.in_IP.s_un_b.s_b[2],
		   DBcfg.in_IP.s_un_b.s_b[3]));
	DB_PRINTF(("DIOCINIT: outside if '%s' hw addr " \
		   "%02x:%02x:%02x:%02x:%02x:%02x\n", req->outName,
		   DBcfg.out_hwaddr.bytes[0], DBcfg.out_hwaddr.bytes[1],
		   DBcfg.out_hwaddr.bytes[2], DBcfg.out_hwaddr.bytes[3],
		   DBcfg.out_hwaddr.bytes[4], DBcfg.out_hwaddr.bytes[5]));
	DB_PRINTF(("DIOCINIT: outside if '%s' ip addr " \
		   "%d.%d.%d.%d\n", req->outName,
		   DBcfg.out_IP.s_un_b.s_b[0],
		   DBcfg.out_IP.s_un_b.s_b[1],
		   DBcfg.out_IP.s_un_b.s_b[2],
		   DBcfg.out_IP.s_un_b.s_b[3]));
	DB_PRINTF(("DIOCINIT: mirror '%s'\n", req->mirrorName));

	/*
	 * All went well so set initialized to yes
	 */
	initialized = YES;
	syslogMessage(SYSL_COLD_START);

	return NO_ERROR;
}

/* ------------------------------------------------------------------------ */

static int
db_ioctl(dev_t dev, u_long cmd, caddr_t addr, int flags, struct proc *p)
{
	u_int s;
	int error = 0;
	static struct timeval stopTime;
	struct timeval curTime;

	if (initialized == 0) {
		if ((cmd != DIOCINIT)  && (cmd != DIOCGLOGF) &&
	    	    (cmd != DIOCSLOGF) && (cmd != DIOCGLOGM) &&
	    	    (cmd != DIOCSLOGM) && (cmd != DIOCGVER)) {
			DB_PRINTF(("DBIOCTL: not initialized\n"));
			return ERROR_NOT_INITILIZED; /* not initialized */
		}
	}

	switch (cmd) {

	case DIOCINIT:		/* Init */
		/*
		 * If drawbridge is running then don't try to reinit
		 */
		if ((flags & FWRITE) == 0)
			return EACCES;
		if (DBstats.running) {
			DB_PRINTF(("DBIOCTL: can't init while drawbridge " \
				   "is running\n"));
			error = ERROR_RUNNING;
			break;
		}
		error = initRequest((initReq *)addr);
		break;

	case DIOCSTART:		/* Start */
		if ((flags & FWRITE) == 0)
			return EACCES;
		if (DBstats.running) {
			DB_PRINTF(("DBIOCTL: drawbridge is already running\n"));
			error = ERROR_RUNNING;
			break;
		}
		/*
                 * put both interfaces into promiscuous mode
                 */
		if (!DBcfg.in_promiscuous) {
			if ((error = db_ifpromisc(DBcfg.in_ifp, 1))) {
				printf("Drawbridge error: couldn't enable " \
					" promiscuous mode for '%s'.\n",
					DBcfg.in_name);
				break;
			}
			DBcfg.in_promiscuous = YES;
		}
		if (!DBcfg.out_promiscuous) {
			if ((error = db_ifpromisc(DBcfg.out_ifp, 1))) {
				printf("Drawbridge error: couldn't enable " \
					" promiscuous mode for '%s'.\n",
					DBcfg.out_name);
				db_ifpromisc(DBcfg.in_ifp, 0);
				DBcfg.in_promiscuous = NO;
				break;
			}
			DBcfg.out_promiscuous = YES;
		}

		getmicrotime(&curTime); 
		if (DBstats.startTime.tv_sec == 0)
			DBstats.startTime = curTime;
		else
			DBstats.startTime.tv_sec += (curTime.tv_sec -
						     stopTime.tv_sec);

		DBstats.running = YES;

		DB_PRINTF(("Drawbridge running...\n"));

		break;

	case DIOCSTOP:		/* Stop */
		if ((flags & FWRITE) == 0)
			return EACCES;
		if (!DBstats.running) {
			DB_PRINTF(("DBIOCTL: drawbridge is not running\n"));
			error = ERROR_NOT_RUNNING; /* not running */
			break;
		}
		DBstats.running = NO;
		getmicrotime(&stopTime); 

		/*
                 * take both interfaces out of promiscuous mode
                 */
		if (DBcfg.in_promiscuous) {
			if ((error = db_ifpromisc(DBcfg.in_ifp, 0))) {
				printf("Drawbridge error: couldn't disable " \
					" promiscuous mode for '%s'.\n",
					DBcfg.in_name);
				break;
			}
			DBcfg.in_promiscuous = NO;
		}
		if (DBcfg.out_promiscuous) {
			if ((error = db_ifpromisc(DBcfg.out_ifp, 0))) {
				printf("Drawbridge error: couldn't disable " \
					" promiscuous mode for '%s'.\n",
					DBcfg.out_name);
				break;
			}
			DBcfg.out_promiscuous = NO;
		}

		DB_PRINTF(("Drawbridge stopped\n"));

		break;

	case DIOCGFLAGS: {	/* Read the flags */
		flagsReq *req = (flagsReq *)addr;

		req->discardMulticast = DBcfg.discardMulticast;
        	req->discardNonIp = DBcfg.discardNonIp;
		req->discardOtherIp = DBcfg.discardOtherIp;
		req->discardSuspectOffset = DBcfg.discardSuspectOffset;
		req->discardFragmentedICMP = DBcfg.discardFragmentedICMP;
		req->discardAttackICMP = DBcfg.discardAttackICMP;
		break;
	}

	case DIOCSFLAGS: {	/* Write the flags */
		flagsReq *req = (flagsReq *)addr;

		if ((flags & FWRITE) == 0)
			return EACCES;
        	DBcfg.discardMulticast = req->discardMulticast;
        	DBcfg.discardNonIp = req->discardNonIp;
		DBcfg.discardOtherIp = req->discardOtherIp;
		DBcfg.discardSuspectOffset = req->discardSuspectOffset;
		DBcfg.discardFragmentedICMP = req->discardFragmentedICMP;
		DBcfg.discardAttackICMP = req->discardAttackICMP;
		break;
	}

	case DIOCGSTATS:	/* Get the stats */
		*((Statistics *)addr) = DBstats;
		break;

	case DIOCCSTATS:	/* Clear the stats */
		if ((flags & FWRITE) == 0)
			return EACCES;
		s = splimp();
		bzero(&DBstats.STATS_CLR_START, STATS_CLR_SIZE);
		splx(s);
		break;

	case DIOCCBRIDGE:	/* Clear the bridge table */
		if ((flags & FWRITE) == 0)
			return EACCES;
		s = splimp();
		bzero(hashTable, sizeof(hashTable));
		DBstats.bridgeEntries = 0;
		splx(s);
		break;

	case DIOCGBRIDGE:	/* Read the bridge table */
		error = bridgeRequest(READ, (bridgeReq *)addr);
		break;

	case DIOCGCLASS:	/* Read from a class table */
		error = portListRequest(READ, (portListReq *)addr, &active);
		break;

	case DIOCSCLASS:	/* Write to a class table */
		if ((flags & FWRITE) == 0)
			return EACCES;
		error = portListRequest(WRITE, (portListReq *)addr, &new);
		break;

	case DIOCGNETWORK:	/* Read from a network table */
		error = addrTreeRequest(READ, (addrTreeReq *)addr);
		break;

	case DIOCSNETWORK:	/* Write to a network table */
		if ((flags & FWRITE) == 0)
			return EACCES;
		error = addrTreeRequest(WRITE, (addrTreeReq *)addr);
		break;

	case DIOCMNETWORK:	/* Manage network tables */
		if ((flags & FWRITE) == 0)
			return EACCES;
		error = manageTablesRequest((manageTablesReq *)addr);
		break;

	case DIOCGOVERRIDE:	/* Read the override table */
		error = overrideTableRequest(READ, (overrideTableReq *)addr,
					     &active);
		break;

	case DIOCSOVERRIDE:	/* Write the override table */
		if ((flags & FWRITE) == 0)
			return EACCES;
		error = overrideTableRequest(WRITE, (overrideTableReq *)addr,
					     &new);
		break;

	case DIOCGREJECT:	/* Read the reject table */
		error = rejaccTableRequest(READ, T_REJTBL,
					   (rejaccTableReq *)addr, &active);
		break;

	case DIOCSREJECT:	/* Write the reject table */
		if ((flags & FWRITE) == 0)
			return EACCES;
		error = rejaccTableRequest(WRITE, T_REJTBL,
					   (rejaccTableReq *)addr, &new);
		break;

	case DIOCGACCEPT:	/* Read the accept address table */
		error = rejaccTableRequest(READ, T_ACCTBL,
					   (rejaccTableReq *)addr, &active);
		break;

	case DIOCSACCEPT:	/* Write the accept address table */
		if ((flags & FWRITE) == 0)
			return EACCES;
		error = rejaccTableRequest(WRITE, T_ACCTBL,
					   (rejaccTableReq *)addr, &new);
		break;

	case DIOCGLOGF:		/* Read the log facility */
		((logFacilityReq *)addr)->facility = DBcfg.logFacility >> 3;
		break;

	case DIOCSLOGF:		/* Write the log facility */
		if ((flags & FWRITE) == 0)
			return EACCES;
		DBcfg.logFacility = ((logFacilityReq *)addr)->facility << 3;
		break;

	case DIOCGLOGM:		/* Read the log mask */
		((logMaskReq *)addr)->mask = DBcfg.logMask;
		break;

	case DIOCSLOGM:		/* Write the log mask */
		if ((flags & FWRITE) == 0)
			return EACCES;
		DBcfg.logMask = ((logMaskReq *)addr)->mask;
		break;

	case DIOCGMGTIF:	/* Read the listen interface */
		((listenModeReq *)addr)->listenMode = DBcfg.listenMode;
		break;

	case DIOCSMGTIF:	/* Set the listen interface */
		if ((flags & FWRITE) == 0)
			return EACCES;
		/* make sure the mode is valid */
		if ((((listenModeReq *)addr)->listenMode & 3)
		    != ((listenModeReq *)addr)->listenMode) {
			return ERROR_INVALID_MODE;
		}
		DBcfg.listenMode = ((listenModeReq *)addr)->listenMode;
		if (DBcfg.listenMode) {
			if ((DBcfg.in_IP.s_addr != 0L) &&
			    (DBcfg.out_IP.s_addr != 0L)) {
				DB_PRINTF(("DIOCINIT: Only one interface may " \
					 "be configured with an IP address\n"));
				return ERROR_TWO_IP;
			} else if ((DBcfg.in_IP.s_addr == 0L) &&
				   (DBcfg.out_IP.s_addr == 0L)) {
				DB_PRINTF(("DIOCINIT: One interface must be " \
					   "configured with an IP address\n"));
				return ERROR_NO_IP;
			}
			if (DBcfg.in_IP.s_addr != 0L) {
				DBcfg.local_IP = DBcfg.in_IP;
				DBcfg.local_hwaddr = DBcfg.in_hwaddr;
				DBcfg.local_ifp = DBcfg.in_ifp;
			} else {
				DBcfg.local_IP = DBcfg.out_IP;
				DBcfg.local_hwaddr = DBcfg.out_hwaddr;
				DBcfg.local_ifp = DBcfg.out_ifp;
			}
		}
		break;

	case DIOCGMMODE:	/* Read the mirror interface mode */
		if (!DBcfg.mirror_ifp)
			return ERROR_MIRROR_UNDEFINED;
		((mirrorModeReq *)addr)->mirrorMode = DBcfg.mirrorMode;
		break;

	case DIOCSMMODE:	/* Set the mirror interface mode */
		if ((flags & FWRITE) == 0)
			return EACCES;
		if (!DBcfg.mirror_ifp)
			return ERROR_MIRROR_UNDEFINED;
		/* make sure the mode is valid */
		switch (((mirrorModeReq *)addr)->mirrorMode) {
		case DB_MIRROR_DISABLE:
		case DB_MIRROR_I:
		case DB_MIRROR_O:
		case DB_MIRROR_B:
		case DB_MIRROR_F:
		case DB_MIRROR_FI:
		case DB_MIRROR_FO:
		case DB_MIRROR_BF:
		case DB_MIRROR_BFI:
		case DB_MIRROR_BFO:
		case DB_MIRROR_IFMASK:
		case DB_MIRROR_MASK:
			DBcfg.mirrorMode = ((mirrorModeReq *)addr)->mirrorMode;
			break;
		default:
			return ERROR_MIRROR_MODE;
		}
		/* setup the mirror variables */
		mirror_i1 = (DBcfg.mirrorMode & DB_MIRROR_I) << 4;
		mirror_o1 = (DBcfg.mirrorMode & DB_MIRROR_O) << 4;
		mirror_i2 = (DBcfg.mirrorMode & (DB_MIRROR_I | DB_MIRROR_F |
				DB_MIRROR_FI | DB_MIRROR_BF |
				DB_MIRROR_BFI)) << 4;
		mirror_o2 = (DBcfg.mirrorMode & (DB_MIRROR_O | DB_MIRROR_F |
				DB_MIRROR_FO | DB_MIRROR_BF |
				DB_MIRROR_BFO)) << 4;
		mirror_b  = (DBcfg.mirrorMode & (DB_MIRROR_I | DB_MIRROR_O |
				DB_MIRROR_B | DB_MIRROR_BF |
				DB_MIRROR_BFI | DB_MIRROR_BFO)) << 4;
		break;

	case DIOCGVER:		/* Read the version of Drawbridge */
		strncpy( ((versionReq *)addr)->version, VERSION, 32 );
		break;

	default:
		DB_PRINTF(("Invalid operation\n"));
		error = ERROR_INVALID_OP;
		break;

	}
	return error;
}

/* ------------------------------------------------------------------------ */

static int db_drvsw_installed = 0;

static
void db_drvinit(void *unused)
{

	if ( ! db_drvsw_installed ) {
#if defined(__FreeBSD__) && __FreeBSD__ >= 4
		cdevsw_add(&db_cdevsw);
#else
		dev_t dev = makedev(CDEV_MAJOR, 0);
		cdevsw_add(&dev, &db_cdevsw, NULL);
#endif
		db_drvsw_installed = 1;
#ifdef DEVFS
		devfs_token = devfs_add_devswf(&db_cdevsw, 0, DV_CHR,
					UID_ROOT, GID_WHEEL, 0600,
					"drawbridge");
#endif
	}
}

SYSINIT(dbdev,SI_SUB_DRIVERS,SI_ORDER_MIDDLE+CDEV_MAJOR,db_drvinit,NULL)

#endif /* DRAWBRIDGE */

/* ------------------------------------------------------------------------ */
