/*
 * Copyright (c) 1993,1994,1995,1997,1998
 *      Texas A&M University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Texas A&M University
 *      and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Developers:
 *     Russell Neeper, David K. Hess, Douglas Lee Schales, David R. Safford
 */

#ifndef _NET_DRAWBRIDGE_H_
#define _NET_DRAWBRIDGE_H_

#include <sys/types.h>
#include <sys/time.h>
#include <syslog.h>

/*==============================================================*/
/*                         Constants                            */
/*==============================================================*/

/* The version string is limited to 32 bytes (including the ending NULL)! */
/* The version number is automatically updated by the upper level make */
#define VERSION			"3.1"

enum { NO, YES };
enum { DISCARD_PKT, ALLOW_PKT, DISCARD_AND_ANSWER_PKT };
enum { ACTIVATE_NEW, CLEAR_ALL, INIT_NEW, MEM_REPORT };
enum { READ, WRITE, MANAGE };
enum { T_DST_IN, T_DST_OUT, T_SRC_IN, T_SRC_OUT,
       U_DST_IN, U_DST_OUT, U_SRC_IN, U_SRC_OUT,
       I_TYP_IN, I_TYP_OUT, T_RST_IN, T_RST_OUT,
       T_ACCTBL, T_REJTBL };
enum { DEFAULT_ACCESS };
enum { NORMAL, INVERT };
enum { FLAG_MULTICAST=0, FLAG_NONIP, FLAG_OTHERIP, FLAG_SUSPECTOFFSET,
       FLAG_FRAGMENTEDICMP, FLAG_ATTACKICMP, MAX_FLAG };
/* The first five of these enum values must not be changed */
enum { DB_DROP_PKT=0, DB_INSIDE_IF=1, DB_OUTSIDE_IF=2, DB_BOTH_IF=3,
       DB_PROCESS_PKT=4, DB_FORWARD_PKT, DB_DROP_AND_ANSWER_PKT };
/* These enum values must not be changed */
enum { DB_MIRROR_DISABLE=0,	/* no pkt mirroring */
       DB_MIRROR_I=0x01,	/* all pkts to/from inside if */
       DB_MIRROR_O=0x02,	/* all pkts to/from outside if */
       DB_MIRROR_B=0x04,	/* pkts bridged */
       DB_MIRROR_F=0x08,	/* pkts filtered */
       DB_MIRROR_FI=0x10,	/* pkts filtered from inside if */
       DB_MIRROR_FO=0x20,	/* pkts filtered from outside if */
       DB_MIRROR_BF=0x40,	/* pkts bridged or filtered */
       DB_MIRROR_BFI=0x80,	/* pkts bridged or filtered from inside if */
       DB_MIRROR_BFO=0x100,	/* pkts bridged or filtered from outside if */
       DB_MIRROR_IFMASK=0x30,
       DB_MIRROR_MASK=0xFFF0
     };

/* -- The maximum number of bridge table entries must also always -- */
/* -- be a power of two.  In this case 2^13 = 8192 entries.       -- */
#define	A_H_BITS		13
#define MAX_BRIDGE_ADDRESSES	(1 << A_H_BITS)			/* 8192 */
#define ADDRESS_HASH_MASK	(MAX_BRIDGE_ADDRESSES - 1)	/* 0x1FFF */

#define LIST_SIZE		32
#define LIST_TABLE_SIZE		(LIST_SIZE * sizeof(PortListEntry))

/*
 * Network packet defines
 */

#define SW_IP_OFFMASK		0xFF1F	/* byte swapped ip offset mask */

/* -- These should be defined in <net/if_llc.h> but are not -- */
#ifndef LLC_SNAPLEN
#define LLC_SNAPLEN		8
#endif
#ifndef llc_snap
#define llc_snap		llc_un.type_snap
#endif

/*
 * Syslog constants. 
 */

#define DEFAULT_SYSL_MASK	0		/* 0x1FFFF for all messages */
#define DEFAULT_SYSL_FACILITY	LOG_LOCAL0

/*
 * Note:  If this enum struct is modified, the corresponding
 * table in drawbridge.c must also be modified.
 */
enum {
	SYSL_UNKNOWN,
	SYSL_COLD_START,	/* cold start */
	SYSL_IN_CLASSD,
	SYSL_OUT_CLASSD,
	SYSL_IN_PORT,
	SYSL_OUT_PORT,		/* only TCP */
	SYSL_IN_TYPE,		/* only srcAddr, dstAddr, and icmp type */
	SYSL_OUT_TYPE,		/* only srcAddr, dstAddr, and icmp type */
	SYSL_IN_REJECT,		/* only protocolNo, srcAddr and dstAddr */
	SYSL_OUT_ACCEPT,	/* only protocolNo, srcAddr and dstAddr */
	SYSL_OUT_OVERRIDE,	/* only TCP */
	SYSL_IN_LENGTH,		/* TCP/UDP-Header maybe not initialized */
	SYSL_OUT_LENGTH,	/* TCP/UDP-Header maybe not initialized */
	SYSL_IN_DOS,		/* denial of service attack packet in */
	SYSL_OUT_DOS,		/* denial of service attack packet out */
	SYSL_IN_PROT,		/* only protocolNo, srcAddr and dstAddr */
	SYSL_OUT_PROT,		/* only protocolNo, srcAddr and dstAddr */
	SYSL_IN_OFFSET,		/* suspect fragment offset */
	SYSL_OUT_OFFSET,	/* suspect fragment offset */
	SYSL_IN_FRAG,		/* fragmented packet */
	SYSL_OUT_FRAG,		/* fragmented packet */
	SYSL_IN_FILTER,		/* no packet information */
	SYSL_OUT_FILTER		/* no packet information */
};

/*
 * Denial of service (DOS) attacks
 */
enum dos_type {
	DOS_SMURF, DOS_PONG
};

/*
 * Error messages.
 */
enum error_msg {
	NO_ERROR=0,
	ERROR_OUT_OF_RANGE=100,	ERROR_UNKNOWN_CLASS,
	ERROR_INVALID_NETWORK,	ERROR_TABLE_FULL,
	ERROR_NET_NOT_FOUND,	ERROR_NO_MEMORY,
	ERROR_INVALID_OP,	ERROR_IF_DOWN,
	ERROR_GET_HW,		ERROR_GET_IP,
	ERROR_INVALID_IF,	ERROR_SAME_IF,
	ERROR_DIFF_IF_TYPE,	ERROR_BAD_IF_TYPE,
	ERROR_INSIDE_DOWN,	ERROR_OUTSIDE_DOWN,
	ERROR_TWO_IP,		ERROR_NO_IP,
	ERROR_INVALID_MODE,	ERROR_NOT_INITILIZED,
	ERROR_RUNNING,		ERROR_NOT_RUNNING,
	ERROR_MIRROR_DOWN,	ERROR_MIRROR_UNDEFINED,
	ERROR_MIRROR_MODE,
	MAX_ERROR_NUM  /* must be last */
};

/*
 * ioctl definitions
 */
#define DIOCINIT	_IOW ('D',  1, initReq)
#define DIOCSTART	_IO  ('D',  2)
#define DIOCSTOP	_IO  ('D',  3)
#define DIOCGFLAGS	_IOR ('D',  4, flagsReq)
#define DIOCSFLAGS	_IOW ('D',  5, flagsReq)
#define DIOCGSTATS	_IOR ('D',  6, Statistics)
#define DIOCCSTATS	_IO  ('D',  7)
#define DIOCGBRIDGE	_IOWR('D',  8, bridgeReq)
#define DIOCCBRIDGE	_IO  ('D',  9)
#define DIOCGCLASS	_IOWR('D', 10, portListReq)
#define DIOCSCLASS	_IOWR('D', 11, portListReq)
#define DIOCGNETWORK	_IOWR('D', 13, addrTreeReq) /* 12 no longer used */
#define DIOCSNETWORK	_IOWR('D', 14, addrTreeReq)
#define DIOCMNETWORK	_IOWR('D', 15, manageTablesReq)
#define DIOCGOVERRIDE	_IOWR('D', 16, overrideTableReq)
#define DIOCSOVERRIDE	_IOWR('D', 17, overrideTableReq)
#define DIOCGREJECT	_IOWR('D', 18, rejaccTableReq)
#define DIOCSREJECT	_IOWR('D', 19, rejaccTableReq)
#define DIOCGACCEPT	_IOWR('D', 20, rejaccTableReq)
#define DIOCSACCEPT	_IOWR('D', 21, rejaccTableReq)
#define DIOCGLOGF	_IOR ('D', 22, logFacilityReq)
#define DIOCSLOGF	_IOW ('D', 23, logFacilityReq)
#define DIOCGLOGM	_IOR ('D', 24, logMaskReq)
#define DIOCSLOGM	_IOW ('D', 25, logMaskReq)
#define DIOCGVER	_IOR ('D', 26, versionReq)
#define DIOCGTCPRESET	_IOR ('D', 27, TCPresetTableReq)
#define DIOCSTCPRESET	_IOW ('D', 28, TCPresetTableReq)
#define DIOCGMGTIF	_IOR ('D', 29, listenModeReq)
#define DIOCSMGTIF	_IOW ('D', 30, listenModeReq)
#define DIOCGMMODE	_IOR ('D', 31, mirrorModeReq)
#define DIOCSMMODE	_IOW ('D', 32, mirrorModeReq)


/*==============================================================*/
/*                           Structures                         */
/*==============================================================*/

typedef union _in_addr {
	u_int32_t s_addr;
	struct {
		u_int8_t s_b[4];
	} s_un_b;
	struct {
		u_int16_t s_w[2];
	} s_un_w;
#define s_b s_un_b.s_b
#define s_w s_un_w.s_w
} in_addr_t;

typedef union _HardwareAddr {
	u_int8_t  bytes[6];
	u_int16_t words[3];
	struct {
		u_int32_t l;
		u_int16_t w;
	} compare;
	struct {
		u_int8_t bytes[6];
	} addr;
} HardwareAddr;

typedef struct _FilterConfig {
	int		logFacility;
	u_int32_t	logMask;
#define CFG_CLR_SIZE	(sizeof(FilterConfig) - sizeof(u_int32_t) - sizeof(int))
#define CFG_CLR_START	in_name[0]
#define MAX_DEV_NAMELEN	16
	char		in_name[MAX_DEV_NAMELEN];
	char		out_name[MAX_DEV_NAMELEN];
	char		mirror_name[MAX_DEV_NAMELEN];
	struct ifnet	*in_ifp;
	struct ifnet	*out_ifp;
	struct ifnet	*mirror_ifp;
	struct ifnet	*local_ifp;
	HardwareAddr	in_hwaddr;
	HardwareAddr	out_hwaddr;
	HardwareAddr	local_hwaddr;
	in_addr_t	in_IP;
	in_addr_t	out_IP;
	in_addr_t	local_IP;
	int		in_promiscuous;
	int		out_promiscuous;
	int		media_type;
	int		listenMode;
	int		mirrorMode;
	int		discardMulticast;
	int		discardNonIp;
	int		discardOtherIp;
	int		discardSuspectOffset;
	int		discardFragmentedICMP;
	int		discardAttackICMP;
} FilterConfig;

typedef struct _Statistics {
	u_short		running;
	struct timeval	startTime;
	u_short		bridgeEntries;
	int		media_type;
#define STATS_CLR_SIZE	( sizeof(Statistics) \
			  - (sizeof(u_short) * 2) \
			  - sizeof(struct timeval) \
			  - sizeof(int) )
#define STATS_CLR_START	droppedIfDown
	u_int32_t	droppedIfDown;
	u_int32_t	droppedIfQfull;
	u_int32_t	droppedNoMbufs;
	u_int64_t	p_insideFiltered;
	u_int64_t	p_outsideFiltered;
	u_int64_t	p_insideRx;
	u_int64_t	p_outsideRx;
	u_int64_t	p_insideTx;
	u_int64_t	p_outsideTx;
	u_int64_t	b_insideFiltered;
	u_int64_t	b_outsideFiltered;
	u_int64_t	b_insideRx;
	u_int64_t	b_outsideRx;
	u_int64_t	b_insideTx;
	u_int64_t	b_outsideTx;

	u_int64_t	f_insideTcpMcast;
	u_int64_t	f_outsideTcpMcast;
	u_int64_t	f_insideUdpMcast;
	u_int64_t	f_outsideUdpMcast;
	u_int64_t	f_insideTcpPort;
	u_int64_t	f_insideTcpPortOverride;
	u_int64_t	f_outsideTcpPort;
	u_int64_t	f_insideUdpPort;
	u_int64_t	f_outsideUdpPort;
	u_int64_t	f_insideIcmpType;
	u_int64_t	f_outsideIcmpType;

	u_int64_t	f_outsideRejectTable;
	u_int64_t	f_outsideSuspectOffset;
	u_int64_t	f_outsideIcmpFrag;
	u_int64_t	f_outsideShortTcpHdr;
	u_int64_t	f_outsideShortUdpHdr;
	u_int64_t	f_outsideSmurfDos;
	u_int64_t	f_outsidePongDos;
	u_int64_t	f_outsideOtherIp;
	u_int64_t	f_outsideNonIp;

	u_int64_t	f_insideAcceptTable;
	u_int64_t	f_insideSuspectOffset;
	u_int64_t	f_insideIcmpFrag;
	u_int64_t	f_insideShortTcpHdr;
	u_int64_t	f_insideShortUdpHdr;
	u_int64_t	f_insideSmurfDos;
	u_int64_t	f_insidePongDos;
	u_int64_t	f_insideOtherIp;
	u_int64_t	f_insideNonIp;

	u_int64_t	insideTCPreset;
	u_int64_t	outsideTCPreset;
} Statistics;

typedef struct _SyslogMessageEntry {
	const char *message;
	unsigned char priority;
} SyslogMessageEntry;

/*
 * Structures for tables.
 */

typedef struct {
        u_int16_t c[256];
} addrTreeLeaf;

typedef struct _addrTreeNode {
        union {
                struct _addrTreeNode *p;
                addrTreeLeaf *pd;
                u_int16_t c;
        } u[256];
        u_int8_t flg[256];
} addrTreeNode;

typedef union {
        addrTreeNode *p;
        addrTreeLeaf *pd;
        u_int16_t c;
} addrTreePtr;

typedef struct _PortListEntry {
	u_int16_t begin;
	u_int16_t end;
} PortListEntry;

typedef struct _RejAccTableEntry {
	in_addr_t network;
	u_int16_t bits;
	u_int16_t flag;
} RejAccTableEntry;

typedef struct _OverrideTableEntry {
	in_addr_t network;
	u_int32_t bits;
	PortListEntry port[LIST_SIZE];
} OverrideTableEntry;

typedef struct {
        u_int32_t		num_classes;
        addrTreeNode		*addrTree;
        PortListEntry		*t_dst_in;
        PortListEntry		*t_dst_out;
        PortListEntry		*t_src_in;
        PortListEntry		*t_src_out;
        PortListEntry		*u_dst_in;
        PortListEntry		*u_dst_out;
        PortListEntry		*u_src_in;
        PortListEntry		*u_src_out;
        PortListEntry		*i_typ_in;
        PortListEntry		*i_typ_out;
	PortListEntry		*t_rst_in;
	PortListEntry		*t_rst_out;
	addrTreeNode		*acceptTree;
	RejAccTableEntry	*acceptTable;
	u_int32_t		accept_bytes;
	addrTreeNode		*rejectTree;
	RejAccTableEntry	*rejectTable;
	u_int32_t		reject_bytes;
	addrTreeNode		*overrideTree;
	OverrideTableEntry	*overrideTable;
	u_int32_t		override_bytes;
	u_int32_t		num_override;
} pointerStruct;

/* Note: it's assumed that this struct will always be 8 bytes long */
typedef struct _HashEntry {
	HardwareAddr address;
	u_int8_t interface;
	u_int8_t valid;
} HashEntry;

/*
 * Dfm request structures for ioctls
 */

typedef struct _initReq {
	char inName[MAX_DEV_NAMELEN];
	char outName[MAX_DEV_NAMELEN];
	char mirrorName[MAX_DEV_NAMELEN];
} initReq;

typedef struct _flagsReq {
	int discardMulticast;
	int discardNonIp;
	int discardOtherIp;
	int discardSuspectOffset;
	int discardFragmentedICMP;
	int discardAttackICMP;
} flagsReq;

typedef struct _bridgeReq {
	HashEntry t;
	u_int32_t index;
} bridgeReq;

typedef struct _portListReq {
	u_int16_t id;
	u_int16_t class;
	PortListEntry *ptr;
	u_int32_t bytes;
} portListReq;   

typedef struct _addrTreeReq {
	int num;
	int newtable;
	in_addr_t addr[128];
	u_int8_t bits[128];
	u_int16_t class[128];
} addrTreeReq;

typedef struct _manageTablesReq {
	int command;
	int val;
} manageTablesReq;

typedef struct _TCPresetTableReq {
	PortListEntry in[LIST_SIZE];
	PortListEntry out[LIST_SIZE];
} TCPresetTableReq;

typedef struct _rejaccTableReq {
	u_int32_t bytes;
	void *ptr;
	u_int32_t addr;
	u_int flag;
} rejaccTableReq;

typedef struct _overrideTableReq {
	u_int32_t bytes;
	void *ptr;
	u_int32_t addr;
	u_int index;
} overrideTableReq;

typedef struct _logFacilityReq {
	int	facility;
} logFacilityReq;

typedef struct _logMaskReq {
	u_int32_t mask;
} logMaskReq;

typedef struct _versionReq {
	u_int8_t version[32];
} versionReq;

typedef struct _listenModeReq {
        int listenMode;
} listenModeReq;

typedef struct _mirrorModeReq {
        int mirrorMode;
} mirrorModeReq;


#if defined(KERNEL) || defined(_KERNEL)
/*==============================================================*/
/*                         Kernel Macros                        */
/*==============================================================*/

#define IS_BROADCAST(address) \
	(*((u_int8_t *) (address)) & 0x01)

#define IS_HW_ADDR_EQU(x1, x2) \
	(((x1)->compare.l == (x2).compare.l) && \
	 ((x1)->compare.w == (x2).compare.w))

#define ARP_TARGET_IP_ADDR(x) \
	(*((u_int32_t *) (((struct ether_arp *) (x))->arp_tpa)))

#define MALLOC_LEAF \
	((addrTreeLeaf *)malloc(sizeof(addrTreeLeaf), M_DEVBUF, M_NOWAIT))

#define MALLOC_NODE \
	((addrTreeNode *)malloc(sizeof(addrTreeNode), M_DEVBUF, M_NOWAIT))

#define MALLOC_CLASS(a) \
	malloc((a) * LIST_TABLE_SIZE, M_DEVBUF, M_NOWAIT);

#ifdef DBDEBUG
#define DB_PRINTF(x)   printf x
#else
#define DB_PRINTF(x)
#endif


/*==============================================================*/
/*                       Kernel Prototypes                      */
/*==============================================================*/

int drawbridge_bridge_lookup __P((HardwareAddr *, struct ifnet *));
int drawbridge_test_pkt __P((struct ifnet **, void *, struct mbuf *));
void drawbridge_response_pkt __P((struct mbuf *, void *, struct mbuf *));
#endif /* ifdef KERNEL */

#endif /* #ifndef _NET_DRAWBRIDGE_H_ */
